"""This module contains client that wrap requests and response to Knoema API"""

import json
import urllib.parse
import urllib.request
import time
import datetime
import hmac
import base64
import hashlib
import random
import string
import io
import os
import knoema.api_definitions as definition
import knoema.api_definitions_sema as definition_sema
import knoema.api_definitions_search as definition_search
from urllib.error import HTTPError

def _random_string(length):
    return ''.join(random.choice(string.ascii_letters) for ii in range(length + 1))

def _string_to_binary(string_data):
    return string_data.encode()

def _crlf():
    return _string_to_binary('\r\n')

def _response_to_json(resp):
    str_response = resp.read().decode('utf-8')

    if resp.status < 200 or resp.status >= 300:
        raise ValueError('Error {} from server:{}', resp.status, str_response)

    #api response can starts with BOM symbol and it break json parser, so have to strip the symbol
    obj_resp = json.loads(str_response.strip('\ufeff'))
    if isinstance(obj_resp, str):
        raise ValueError(obj_resp)

    return obj_resp

class ApiClient:
    """This is client that wrap requests and response to Knoema API"""

    def __init__(self, host, appid=None, appsecret=None):
        splitted = urllib.parse.urlsplit(host)
        self._host = splitted.netloc.strip()
        if not self._host:
            self._host = splitted.path.strip()
        self._schema = splitted.scheme
        if not self._schema:
            self._schema = 'http'

        self._appid = appid
        self._appsecret = appsecret
        self._opener = urllib.request.build_opener(urllib.request.HTTPCookieProcessor)

        self._search_config = None

    def _get_url(self, apipath):
        return urllib.parse.urlunsplit((self._schema, self._host, apipath, '', ''))

    def _get_request_headers(self):

        if not self._appid or not self._appsecret:
            return {
                'Content-Type' : 'application/json',
                'Accept': 'application/json'
                }

        key = datetime.datetime.utcnow().strftime('%d-%m-%y-%H').encode()
        hashed = hmac.new(key, self._appsecret.encode(), hashlib.sha1)
        secrethash = base64.b64encode(hashed.digest()).decode('utf-8')
        auth = 'Knoema {}:{}:1.2'.format(self._appid, secrethash)

        return {
            'Content-Type' : 'application/json',
            'Accept': 'application/json',
            'Authorization' : auth
            }

    def _api_get(self, obj, apipath, query=None):

        url = self._get_url(apipath)
        if query:
            url = '{}?{}'.format(url, query)

        headers = self._get_request_headers()
        req = urllib.request.Request(url, headers=headers)
        resp = self._opener.open(req)
        return obj(_response_to_json(resp))

    def _api_post(self, responseobj, apipath, requestobj):

        json_data = requestobj.save_to_json()
        
        return self._api_post_json(responseobj, apipath, json_data)

    def _api_post_json(self, responseobj, apipath, requestjson):

        url = self._get_url(apipath)

        binary_data = requestjson.encode()

        headers = self._get_request_headers()
        req = urllib.request.Request(url, binary_data, headers)
        resp = self._opener.open(req)
        return responseobj(_response_to_json(resp))

    def check_correct_host(self):
        pass 

    def get_dataset(self, datasetid):
        """The method is getting information about dataset byt it's id"""

        path = '/api/1.0/meta/dataset/{}'
        return self._api_get(definition.Dataset, path.format(datasetid))

    def get_dataset_meta(self, datasetid):
        path = '/api/1.0/meta/dataset/{}'
        return self._api_get(definition.DatasetMetadata, path.format(datasetid))

    def get_dimension(self, dataset, dimension):
        """The method is getting information about dimension with items"""

        path = '/api/1.0/meta/dataset/{}/dimension/{}'
        return self._api_get(definition.Dimension, path.format(dataset, dimension))

    def get_daterange(self, dataset):
        """The method is getting information about date range of dataset"""

        path = '/api/1.0/meta/dataset/{}/daterange'
        return self._api_get(definition.DateRange, path.format(dataset))

    def get_data(self, pivotrequest):
        """The method is getting data by pivot request"""

        path = '/api/1.0/data/pivot/'
        return self._api_post(definition.PivotResponse, path, pivotrequest)

    def get_data_by_json(self, pivotrequest_json):
        """The method is getting data by pivot request (json)"""

        path = '/api/1.0/data/pivot/'
        return self._api_post_json(definition.PivotResponse, path, pivotrequest_json)

    def get_dataset_data(self, dataset_id, filters):
        """The method is getting JSON by URL and parses it to specified object"""
        try:
            return self._api_post(definition.detect_data_response, '/api/2.0/data?datasetId={}'.format(dataset_id), filters)
        except HTTPError as ex:
            if ex.code == 400:
                raise ValueError(ex.read().decode('utf-8'))
            else:
                raise

    def get_data_raw(self, request, metadata_only = False):
        """The method is getting data by raw request"""
        path = '/api/1.2/data/raw/' + ('?metadataOnly=true' if metadata_only else '')
        res = self._api_post(definition.RawDataResponse, path, request)
        token = res.continuation_token
        while token is not None:
           res2 = self.get_data_raw_with_token(token, metadata_only)
           res.series += res2.series
           token = res2.continuation_token 
        return res

    def get_data_raw_with_token(self, token, metadata_only = False):
        path = '/api/1.0/data/raw/?continuationToken={0}' + ('&metadataOnly=true' if metadata_only else '')
        return self._api_get(definition.RawDataResponse, path.format(token))

    def get_mnemonics(self, mnemonics, transform, frequency):
        """The method get series by mnemonics"""
        path = '/api/1.0/data/mnemonics?mnemonics={0}'
        if transform:
            path += '&transform=' + transform
        if frequency:
            path += '&frequency=' + frequency
        return self._api_get(definition.MnemonicsResponseList, path.format(mnemonics))

    def get_details(self, request):
        """The method is getting data details by request"""
        path = '/api/1.1/data/details/'
        return self._api_post(definition.DetailsResponse, path, request)

    def get_company_info(self, ticker):
        """The method get company data"""

        path = 'api/1.0/sema/{0}'
        return self._api_get(definition_sema.CompanyInt, path.format(ticker))

    def get_indicator_info(self, path):
        path = 'api/1.0/sema/{0}'.format(path)
        url = self._get_url(path)

        headers = self._get_request_headers()
        req = urllib.request.Request(url, headers=headers)
        resp = self._opener.open(req)
        return _response_to_json(resp)

    def search(self, query):
        if self._search_config == None:
            path = '/api/1.0/search/config'
            self._search_config = self._api_get(definition_search.SearchConfig, path)

        headers = self._get_request_headers()
        url = self._search_config.build_search_url(query)
        req = urllib.request.Request(url, headers=headers)
        req = urllib.request.Request(url)
        resp = self._opener.open(req)

        return definition_search.SearchResultsInt(_response_to_json(resp))

    def upload_file(self, file):
        """The method is posting file to the remote server"""

        url = self._get_url('/api/1.0/upload/post')

        fcontent = FileContent(file)
        binary_data = fcontent.get_binary()

        headers = self._get_request_headers()
        req = urllib.request.Request(url, binary_data, headers)
        req.add_header('Content-type', fcontent.get_content_type())
        req.add_header('Content-length', len(binary_data))
        resp = urllib.request.urlopen(req)   

        return definition.UploadPostResponse(_response_to_json(resp))

    def upload_verify(self, file_location, dataset=None):
        """This method is verifiing posted file on server"""

        path = '/api/1.0/upload/verify'
        query = 'doNotGenerateAdvanceReport=true&filePath={}'.format(file_location)
        if dataset:
            query = 'doNotGenerateAdvanceReport=true&filePath={}&datasetId={}'.format(file_location, dataset)

        return self._api_get(definition.UploadVerifyResponse, path, query)

    def upload_submit(self, upload_request):
        """The method is submitting dataset upload"""

        path = '/api/1.0/upload/save'
        return self._api_post(definition.DatasetUploadResponse, path, upload_request)

    def upload_status(self, upload_id):
        """The method is checking status of uploaded dataset"""

        path = '/api/1.0/upload/status'
        query = 'id={}'.format(upload_id)
        return self._api_get(definition.DatasetUploadStatusResponse, path, query)

    def upload(self, file_path, dataset=None, public=False, name = None):
        """Use this function to upload data to Knoema dataset."""

        upload_status = self.upload_file(file_path)
        err_msg = 'Dataset has not been uploaded to the remote host'
        if not upload_status.successful:
            msg = '{}, because of the following error: {}'.format(err_msg, upload_status.error)
            raise ValueError(msg)

        err_msg = 'File has not been verified'
        upload_ver_status = self.upload_verify(upload_status.properties.location, dataset)
        if not upload_ver_status.successful:
            ver_err = '\r\n'.join(upload_ver_status.errors)
            msg = '{}, because of the following error(s): {}'.format(err_msg, ver_err)
            raise ValueError(msg)

        ds_upload = definition.DatasetUpload(upload_ver_status, upload_status, dataset, public, name)
        ds_upload_submit_result = self.upload_submit(ds_upload)
        err_msg = 'Dataset has not been saved to the database'
        if ds_upload_submit_result.status == 'failed':
            ver_err = '\r\n'.join(ds_upload_submit_result.errors)
            msg = '{}, because of the following error(s): {}'.format(err_msg, ver_err)
            raise ValueError(msg)

        ds_upload_result = None
        while True:
            ds_upload_result = self.upload_status(ds_upload_submit_result.submit_id)
            if ds_upload_result.status == 'pending' or ds_upload_result.status == 'processing':
                time.sleep(5)
            else:
                break

        if ds_upload_result.status != 'successful':
            ver_err = '\r\n'.join(ds_upload_result.errors)
            msg = '{}, because of the following error(s): {}'.format(err_msg, ver_err)
            raise ValueError(msg)

        return ds_upload_result.dataset

    def delete(self, dataset):
        """The method is deleting dataset by it's id"""

        url = self._get_url('/api/1.0/meta/dataset/{}/delete'.format(dataset))

        json_data = ''
        binary_data = json_data.encode()

        headers = self._get_request_headers()
        req = urllib.request.Request(url, binary_data, headers)
        resp = urllib.request.urlopen(req)   
        str_response = resp.read().decode('utf-8')
        if str_response != '"successful"' or resp.status < 200 or resp.status >= 300:
            msg = 'Dataset has not been deleted, because of the following error(s): {}'.format(str_response)
            raise ValueError(msg)

    def verify(self, dataset, publication_date, source, refernce_url):
        """The method is verifying dataset by it's id"""

        path = '/api/1.0/meta/verifydataset'
        req = definition.DatasetVerifyRequest(dataset, publication_date, source, refernce_url)
        result = self._api_post(definition.DatasetVerifyResponse, path, req)
        if result.status == 'failed':
            ver_err = '\r\n'.join(result.errors)
            msg = 'Dataset has not been verified, because of the following error(s): {}'.format(ver_err)
            raise ValueError(msg)


class FileContent(object):
    """Accumulate the data to be used when posting a form."""

    def __init__(self, file):
        self.file_name = os.path.basename(file)
        self.body = open(file, mode='rb').read()
        self.boundary = _random_string(30)

    def get_content_type(self):
        """Return a content type"""
        return 'multipart/form-data; boundary="{}"'.format(self.boundary)

    def get_binary(self):
        """Return a binary buffer containing the file content"""

        content_disp = 'Content-Disposition: form-data; name="file"; filename="{}"'

        stream = io.BytesIO()
        stream.write(_string_to_binary('--{}'.format(self.boundary)))
        stream.write(_crlf())
        stream.write(_string_to_binary(content_disp.format(self.file_name)))
        stream.write(_crlf())
        stream.write(_crlf())
        stream.write(self.body)
        stream.write(_crlf())
        stream.write(_string_to_binary('--{}--'.format(self.boundary)))
        stream.write(_crlf())

        return stream.getvalue()



"""This module contains metadata definitions for Knoema API for semantic atlas"""

from knoema.data_reader import TransformationDataReader
import urllib.parse

class SearchConfig(object):

    def __init__(self, data):
        self.host = data['host']
        self.search_host = data['searchHost']
        self.lang = data['lang']
        self.access_token = data['accessToken']
        self.session_id = data['sessionId']
        self.community_id = data['communityId']

    def build_search_url(self, query):
        quoted_query = urllib.parse.quote(query)
        url = 'https://{}/api/1.0/search?query={}&scope=instant,semantic&count=8&version=4&host={}&lang={}&sessionId={}&access_token={}'
        url = url.format(self.search_host, quoted_query, self.host, self.lang, self.session_id, self.access_token or "")

        if self.community_id:
           url += '&communityId={}'.format(self.community_id)

        return url


class Dimension(object):

    def __init__(self, data):
        self.name = data['dimension']
        self.key = data['key']
        self.value = data['name']

class SearchResults(object):

    def __init__(self, results, client):
        self.instant = None
        if results.instant != None:
            if results.instant.type == 'ConceptBind':
                self.instant = CompanyResult(results.instant)

        self.series = []
        for series in results.series:
            self.series.append(TimeseriesSearchResult(series, client))


class SearchResult(object):

    def __init__(self, result):
        self.type = result.type


class TimeseriesSearchResult(SearchResult):

    def __init__(self, result, client):
        super().__init__(result)

        self._client = client

        self.title = result.title
        self.dataset = result.dataset
        self.frequency = result.frequency
        self.start_date = result.start_date
        self.end_date = result.end_date
        self.dimensions = []
        self._dim_values = {}
        for dim in result.dimensions:
            self._dim_values[dim.name] = dim.value
            self.dimensions.append(dim)

    def get(self, transform = None):
        ds = self._client.get_dataset(self.dataset)

        reader =  TransformationDataReader(self._client, self._dim_values, transform, self.frequency)
        reader.dataset = ds
        
        return reader.get_pandasframe()


class CompanyResult(SearchResult):

    def __init__(self, company):
        super().__init__(company)

        self.name = company.name
        self.id = company.id
        self.company = None


class SearchResultsInt(object):

    def __init__(self, data):

        self.instant = None
        self.series = []

        if data['totalItems'] == 0:
            return

        items = data['items']
        for item in items:
            if item['type'] == 'ConceptBind':
                if item['concepts'][0]['conceptType'] == 'Company':
                    self.instant = CompanyResultInt(item)

            if item['type'] == 'TimeSeries':
                self.series.append(TimeseriesSearchResultInt(item))


class SearchResultInt(object):

    def __init__(self, data):
        self.type = data['type']

class TimeseriesSearchResultInt(SearchResultInt):

    def __init__(self, data):
        super().__init__(data)

        self.title = data['title']
        self.dataset = data['dataset']['id']
        self.frequency = data['frequency']
        self.start_date = data['startDate']
        self.end_date = data['endDate']
        self.dimensions = []
        self._dim_values = {}
        for dim in data['dimensions']:
            dimension = Dimension(dim)
            self._dim_values[dimension.name] = dimension.value
            self.dimensions.append(dimension)


class CompanyResultInt(SearchResultInt):

    def __init__(self, data):
        super().__init__(data)

        self.name = data['concepts'][0]['title']
        self.id = data['concepts'][0]['id']



"""This is test module for knoema client with test credentials"""

import unittest
import knoema
import pandas

class TestKnoemaClient(unittest.TestCase):
    """This is class with knoema client unit tests with test credentials"""

    base_host = 'knoema.com'

    def setUp(self):
        apicfg = knoema.ApiConfig()
        apicfg.host = self.base_host
        apicfg.app_id = 'FzOYqDg'
        apicfg.app_secret = 'SPrvmY8eGRcGA'

    def test_getdata_singleseries_by_member_id(self):
        """The method is testing getting single series by dimension member ids"""

        data_frame = knoema.get('xmhdwqf', company='c1', indicator='ind_a')
        self.assertEqual(data_frame.shape[0], 11)
        self.assertEqual(data_frame.shape[1], 1)

        self.assertEqual(['Company', 'Indicator', 'Frequency'], data_frame.columns.names)

        indx = data_frame.first_valid_index()
        sname = ('BOX', 'Annual', 'A')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 85.50)

        indx = data_frame.index[10]
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 15.62)

    def test_getdata_multiseries_by_member_id(self):
        """The method is testing getting multiple series by dimension member ids"""

        data_frame = knoema.get('xmhdwqf', company='c1;c2', indicator='ind_m;ind_a')
        self.assertEqual(data_frame.shape[0], 56)
        self.assertEqual(data_frame.shape[1], 4)

        self.assertEqual(['Company', 'Indicator', 'Frequency'], data_frame.columns.names)

        indx = data_frame.index[7]
        sname = ('BOX', 'Monthly', 'M')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 23.08)

        indx = data_frame.index[55]
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 19.71)

    def test_getdata_multiseries_by_member_name(self):
        """The method is testing getting data by dimension member names"""

        company_names = 'BOX;UBER'
        indicator_names = 'Monthly;Annual'
        data_frame = knoema.get('xmhdwqf', company=company_names, indicator=indicator_names)
        self.assertEqual(data_frame.shape[0], 56)
        self.assertEqual(data_frame.shape[1], 4)

        self.assertEqual(['Company', 'Indicator', 'Frequency'], data_frame.columns.names)

        indx = data_frame.index[7]
        sname = ('BOX', 'Monthly', 'M')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 23.08)

        indx = data_frame.index[55]
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 19.71)

    def test_getdata_multiseries_by_member_id_range(self):
        """The method is testing getting multiple series by dimension member ids and time range"""
        data_frame = knoema.get('xmhdwqf', company='c1;c2', indicator='ind_a', timerange='2017-2019')

        self.assertEqual(data_frame.shape[0], 11)
        self.assertEqual(data_frame.shape[1], 2)

        indx = data_frame.first_valid_index()
        sname = ('UBER', 'Annual', 'A')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 53.03)

        indx = data_frame.last_valid_index()
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 99.15)
   
    def test_getdata_singleseries_difffrequencies_by_member_id(self):
        """The method is testing getting single series on different frequencies by dimension member ids"""

        data_frame = knoema.get('xmhdwqf', company='c1', indicator='ind_multi')
        self.assertEqual(data_frame.shape[1], 3)

        indx = data_frame.first_valid_index()
        sname = ('BOX', 'Multi', 'A')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 60.24)

        value = data_frame.at[pandas.to_datetime('2018-01-01'), sname]
        self.assertEqual(value, 80.56)

        indx = data_frame.first_valid_index()
        sname = ('BOX', 'Multi', 'Q')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 47.82)

        value = data_frame.at[pandas.to_datetime('2017-01-01'), sname]
        self.assertEqual(value, 50.28)

    def test_getdata_multiseries_singlefrequency_by_member_id(self):
        """The method is testing getting mulitple series with one frequency by dimension member ids"""

        data_frame = knoema.get('xmhdwqf', company='c2', indicator='ind_multi', frequency='M')
        self.assertEqual(data_frame.shape[1], 1)

        sname = ('UBER', 'Multi', 'M')
        value = data_frame.at[pandas.to_datetime('2017-01-01'), sname]
        self.assertEqual(value, 27.91)

    def test_getdata_multiseries_multifrequency_by_member_id(self):
        """The method is testing getting mulitple series queriing mulitple frequencies by dimension member ids"""

        data_frame = knoema.get('xmhdwqf', company='c1;c2', indicator='ind_a;ind_multi', frequency='A;M')
        self.assertEqual(data_frame.shape[1], 6)

        sname = ('BOX', 'Annual', 'A')
        value = data_frame.at[pandas.to_datetime('2013-01-01'), sname]
        self.assertEqual(value, 77.93)

    def test_getdata_multiseries_multifrequency_by_member_id_range(self):
        """The method is testing getting mulitple series queriing mulitple frequencies by dimension member ids with time range"""

        data_frame = knoema.get('xmhdwqf', company='c1;c2', indicator='ind_a;ind_multi', frequency='A;M', timerange='2017M1-2020M12')
        self.assertEqual(data_frame.shape[1], 6)

        sname = ('BOX', 'Multi', 'A')
        value = data_frame.at[pandas.to_datetime('2018-01-01'), sname]
        self.assertEqual(value, 80.56)

    def test_none_dataset(self):
        """The method is testing if dataset set up as None"""

        with self.assertRaises(ValueError) as context:
            knoema.get(None)

        self.assertTrue('Dataset id is not specified' in str(context.exception))       

    def test_wrong_dimension(self):
        """The method is testing if there is wrong dimension name is specified"""
   
        with self.assertRaises(ValueError) as context:
            knoema.get('IMFWEO2017Apr', indicator='LP;NGDP')

        self.assertTrue('Dimension with id or name indicator is not found' in str(context.exception))

    def test_wrong_dimension_with_transform(self):
        """The method is testing if there is wrong dimension name is specified"""
   
        with self.assertRaises(ValueError) as context:
            knoema.get('IMFWEO2017Apr', indicator='LP;NGDP', transform='PCH')

        self.assertTrue('Dimension with id or name indicator is not found' in str(context.exception))

    def test_empty_dimension_selection(self):
        """The method is testing if there are no elements in dimension selection"""

        with self.assertRaises(ValueError) as context:
            knoema.get('IMFWEO2017Apr', country='', subject='LP;NGDP')

        self.assertTrue('Selection for dimension Country is empty' in str(context.exception))

    def test_empty_dimension_selection_with_transform(self):
        """The method is testing if there are no elements in dimension selection"""

        with self.assertRaises(ValueError) as context:
            knoema.get('IMFWEO2017Apr', country='', subject='LP;NGDP', transform='PCH')

        self.assertTrue('Selection for dimension Country is empty' in str(context.exception))


    def test_wrong_dimension_selection(self):
        """The method is testing if there are incorrect in dimension selection"""

        with self.assertRaises(ValueError) as context:
            knoema.get('xmhdwqf', company='c1;c2', indicator='ind_a;ind_multi;ind')

        self.assertTrue('Selection for dimension Indicator contains invalid elements' in str(context.exception))

    def test_wrong_dimension_selection_transform(self):
        """The method is testing if there are incorrect in dimension selection"""

        with self.assertRaises(ValueError) as context:
            knoema.get('xmhdwqf', company='c1;c2', indicator='ind_a;ind_multi;ind', frequency='A', transform='sum')

        self.assertTrue('Selection for dimension Indicator contains invalid elements' in str(context.exception))

    def test_get_data_from_flat_dataset_with_datecolumn(self):
        """The method is testing load data from flat dataset with specifying datecolumn"""

        data_frame = knoema.get('bjxchy', country='Albania', measure='Original Principal Amount ($)', datecolumn='Effective Date (Most Recent)', timerange='2010-2015', frequency='A')
        self.assertEqual(data_frame.shape[0], 5)
        self.assertEqual(data_frame.shape[1], 5)

        sname = ('Albania', 'MINISTRY OF FINANCE', 'Albania', 'FSL', 'Repaying', 'Sum(Original Principal Amount ($))', 'A')
        value = data_frame.at[pandas.to_datetime('2013-01-01'), sname]
        self.assertEqual(value, 40000000.0)

    def test_incorrect_dataset_id(self):
        """The method is testing if dataset id set up incorrectly"""

        with self.assertRaises(ValueError) as context:
            knoema.get('incorrect_id', somedim='val1;val2')

        self.assertTrue("Requested dataset doesn't exist or you don't have access to it." in str(context.exception))

    def test_getdata_multiseries_by_member_key(self):
        """The method is testing getting multiple series by dimension member keys"""

        data_frame = knoema.get('xmhdwqf', company='1000000;1000010', indicator='1000020;1000050')
        self.assertEqual(data_frame.shape[0], 56)
        self.assertEqual(data_frame.shape[1], 4)

        self.assertEqual(['Company', 'Indicator', 'Frequency'], data_frame.columns.names)

        indx = data_frame.index[7]
        sname = ('BOX', 'Monthly', 'M')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 23.08)

        indx = data_frame.index[55]
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 19.71)

    def test_get_data_from_dataset_with_multiword_dimnames_and_metadata_and_mnemomics(self):
        """The method is testing load data from regular dataset with dimenions that have multi word names include metadata and mnemonics"""

        data_frame, metadata = knoema.get('eqohmpb', True, **{'Country': '1000000',
                                                    'Indicator': '1000010'})

        self.assertEqual(data_frame.shape[0], 22)
        self.assertEqual(data_frame.shape[1], 6)
        self.assertEqual(metadata.shape[0], 5)
        self.assertEqual(metadata.shape[1], 6)
        self.assertEqual(['Country', 'Indicator', 'Frequency'], data_frame.columns.names)
        self.assertEqual(['Country', 'Indicator', 'Frequency'], metadata.columns.names)

        sname = ('Afghanistan', 'Gross domestic product, current prices', 'A')

        indx = data_frame.first_valid_index()
        value = data_frame.at[indx, sname]
        self.assertAlmostEqual(value, 0.0)

        indx = metadata.first_valid_index()
        value = metadata.at[indx, sname]
        self.assertAlmostEqual(value, '512')

        self.assertAlmostEqual(metadata.at['Unit', sname], 'Number')
        self.assertAlmostEqual(metadata.at['Scale', sname], 1.0)
        self.assertAlmostEqual(metadata.at['Mnemonics', sname], '512NGDP_A_in_test_dataset')

    def test_weekly_frequency(self):
        """The method is testing load data from regular dataset by weekly frequency"""
   
        data = knoema.get('xmhdwqf', company='BOX', indicator='Weekly', frequency='W')
        sname = ('BOX', 'Weekly', 'W')
        value = data.at[pandas.to_datetime('2018-12-31'), sname]
        self.assertEqual(value, 32.37)
        value = data.at[pandas.to_datetime('2019-12-30'), sname]
        self.assertEqual(value, 83.73)

    def test_incorrect_host_knoema_get(self):
        """The method is negative test on get series from dataset with incorrect host"""

        with self.assertRaises(ValueError) as context:
            apicfg = knoema.ApiConfig()
            apicfg.host = 'knoema_incorect.com'
            _ = knoema.get('IMFWEO2017Apr', country='914', subject='ngdp')
        self.assertTrue("The specified host knoema_incorect.com does not exist" in str(context.exception))
 
    def test_search_by_mnemonics_data(self):
        """The method is testing searching by mnemonics"""

        data_frame = knoema.get('eqohmpb', mnemonics='512NGDP_A_in_test_dataset')
        self.assertEqual(data_frame.shape[1], 1)
        sname = ('512NGDP_A_in_test_dataset')
        indx = data_frame.first_valid_index()
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 0)

    def test_search_by_mnemonics_with_metadata(self):
        """The method is testing searching by mnemonics with metadata"""

        data_frame, metadata = knoema.get('eqohmpb', True, mnemonics='512NGDP_A_in_test_dataset')
        self.assertEqual(data_frame.shape[1], 1)
        sname = ('512NGDP_A_in_test_dataset')
        indx = data_frame.first_valid_index()
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 0)
        self.assertEqual(metadata.shape[1], 1)
        self.assertEqual(metadata.shape[0], 5)
        self.assertEqual(metadata.at['Country Id',sname],'512')
        self.assertEqual(metadata.at['Indicator Id',sname],'NGDP')
        self.assertEqual(metadata.at['Unit',sname],'Number')
        self.assertEqual(metadata.at['Mnemonics',sname],'512NGDP_A_in_test_dataset')

    def test_search_by_mnemonics_data_by_all_datasets(self):
        """The method is testing searching by mnemonics by all dataset and returns data"""
        data_frame = knoema.get(mnemonics='512NGDP_A_in_test_dataset;512NGDP')
        self.assertEqual(data_frame.shape[1], 2)
        sname = ('512NGDP_A_in_test_dataset')
        indx = data_frame.first_valid_index()
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 0)

    def test_search_by_mnemonics_with_metadata_by_all_datasets(self):
        """The method is testing searching by mnemonics by all dataset and returns data and metadata"""
        data_frame, metadata = knoema.get(None, True, mnemonics='512NGDP_A_in_test_dataset;512NGDP')
        self.assertEqual(data_frame.shape[1], 2)
        sname = ('512NGDP_A_in_test_dataset')
        indx = data_frame.first_valid_index()
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 0)
        self.assertEqual(metadata.at['Country Id',sname],'512')
        self.assertEqual(metadata.at['Indicator Id',sname],'NGDP')
        self.assertEqual(metadata.at['Unit',sname],'Number')
        self.assertEqual(metadata.at['Mnemonics',sname],'512NGDP_A_in_test_dataset')

    def test_search_by_mnemonics_with_metadata_by_all_datasets_transform(self):
        """The method is testing searching by mnemonics by all dataset and returns data and metadata"""
        data_frame, metadata = knoema.get(None, True, mnemonics='512NGDP_R;512NGDP', transform='PCH', frequency='A')
        self.assertEqual(data_frame.shape[1], 2)
        sname = ('512NGDP_R')
        indx = data_frame.first_valid_index()
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 8.69275111845198)
        self.assertEqual(metadata.at['Country Id',sname],'512')
        self.assertEqual(metadata.at['Indicator Id',sname],'NGDP_R')
        self.assertEqual(metadata.at['Unit',sname],'%')
        self.assertEqual(metadata.at['Mnemonics',sname],'512NGDP_R')
    
    def test_get_all_series_from_dataset(self):
        """The method is testing getting all series from dataset"""
        data_frame= knoema.get('dzlnsee')
        self.assertEqual(data_frame.shape[1], 4)
        self.assertEqual(['Company', 'Indicator', 'Frequency'], data_frame.columns.names)

        indx = data_frame.first_valid_index()
        sname = ('BOX', 'Monthly', 'M')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 23.08)

    def test_with_empty_attributes_in_the_dimensions(self):
        data, metadata = knoema.get('IMFWEO2017Oct', include_metadata=True, country=['914','512'], subject='lp')
        self.assertEqual(data.shape[0], 43)
        self.assertEqual(data.shape[1], 2)
        self.assertEqual(['Country', 'Subject', 'Frequency'], data.columns.names)
        self.assertEqual(metadata.shape[0], 8)
        self.assertEqual(metadata.shape[1], 2)
        self.assertEqual(['Country', 'Subject', 'Frequency'], metadata.columns.names)
        indx = data.first_valid_index()
        sname = ('Albania', 'Population (Persons)', 'A')
        value = data.at[indx, sname]
        self.assertEqual(value, 2.762)
        
        self.assertEqual(metadata.at['Country Id',sname],'914')
        self.assertEqual(metadata.at['Subject Id',sname],'LP')
        self.assertEqual(metadata.at['Subject SubjectDescription',sname],None)
        self.assertEqual(metadata.at['Subject SubjectNotes',sname],None)
        self.assertEqual(metadata.at['Subject Relevant',sname],None)
        self.assertEqual(metadata.at['Unit',sname],'Persons (Millions)')
        self.assertEqual(metadata.at['Mnemonics',sname],None)

        indx = data.last_valid_index()
        value = data.at[indx, sname]
        self.assertEqual(value, 2.856)
        
    def test_with_empty_attributes_in_the_dimensions_transform(self):
        data, metadata = knoema.get('IMFWEO2017Oct', include_metadata=True, country=['914','512'], subject='lp', frequency='A')
        self.assertEqual(data.shape[0], 43)
        self.assertEqual(data.shape[1], 2)
        self.assertEqual(['Country', 'Subject', 'Frequency'], data.columns.names)
        self.assertEqual(metadata.shape[0], 8)
        self.assertEqual(metadata.shape[1], 2)
        self.assertEqual(['Country', 'Subject', 'Frequency'], metadata.columns.names)
        indx = data.first_valid_index()
        sname = ('Albania', 'Population (Persons)', 'A')
        value = data.at[indx, sname]
        self.assertEqual(value, 2.762)
        
        self.assertEqual(metadata.at['Country Id',sname],'914')
        self.assertEqual(metadata.at['Subject Id',sname],'LP')
        self.assertEqual(metadata.at['Subject SubjectDescription',sname],None)
        self.assertEqual(metadata.at['Subject SubjectNotes',sname],None)
        self.assertEqual(metadata.at['Subject Relevant',sname],None)
        self.assertEqual(metadata.at['Unit',sname],'Persons (Millions)')
        self.assertEqual(metadata.at['Mnemonics',sname],None)

        indx = data.last_valid_index()
        value = data.at[indx, sname]
        self.assertEqual(value, 2.856)

        
    def test_getdata_from_private_community(self):
        """The method is testing getting data from private community"""

        apicfgCommunity = knoema.ApiConfig()
        apicfgCommunity.host = 'teryllol.' + self.base_host
        apicfgCommunity.app_id = 's81oiSY'
        apicfgCommunity.app_secret='g4lKmIOPE2R4w'

        data_frame = knoema.get('qfsneof', country='USA', series='NY.GDP.MKTP.KD.ZG')
        self.assertEqual(data_frame.shape[0], 60)
        self.assertEqual(data_frame.shape[1], 1)

        self.assertEqual(['Country', 'Series', 'Frequency'], data_frame.columns.names)

        indx = data_frame.first_valid_index()
        sname = ('United States', 'GDP growth (annual %)', 'A')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 2.29999999999968)

        indx = data_frame.index[57]
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 2.99646435222829)


    def test_getdata_custom_separator(self):
        """The method is testing getting data with custom separator"""

        data_frame = knoema.get('IMFWEO2019Oct',
            country='Albania',
            subject='Gross domestic product, constant prices (Percent change)|Gross domestic product per capita, constant prices (Purchasing power parity; 2011 international dollar)',
            separator='|')
        self.assertEqual(data_frame.shape[0], 45)
        self.assertEqual(data_frame.shape[1], 2)

        self.assertEqual(['Country', 'Subject', 'Frequency'], data_frame.columns.names)

        indx = data_frame.first_valid_index()
        sname = ('Albania', 'Gross domestic product per capita, constant prices (Purchasing power parity; 2011 international dollar)', 'A')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 4832.599)

    def test_getdata_default_web_separator(self):
        """The method is testing getting data with custom separator"""

        data_frame = knoema.get('IMFWEO2019Oct',
            country='Albania',
            subject='NGDP_RPCH,NGDPRPPPPC',
            separator=',')
        self.assertEqual(data_frame.shape[0], 45)
        self.assertEqual(data_frame.shape[1], 2)

        self.assertEqual(['Country', 'Subject', 'Frequency'], data_frame.columns.names)

        indx = data_frame.first_valid_index()
        sname = ('Albania', 'Gross domestic product per capita, constant prices (Purchasing power parity; 2011 international dollar)', 'A')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 4832.599)

    def test_getdata_region_dim_region_id(self):
        """The method is testing getting data from dataset with region dimention by region id"""

        data_frame = knoema.get('IMFWEO2019Oct',
            country = 'AF',
            subject = ['Gross domestic product, constant prices (Percent change)']
        )

        self.assertEqual(['Country', 'Subject', 'Frequency'], data_frame.columns.names)
        self.assertEqual('Afghanistan', data_frame.columns.values[0][0])

        indx = data_frame.first_valid_index()
        sname = ('Afghanistan', 'Gross domestic product, constant prices (Percent change)', 'A')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 8.692)

    def test_getdata_region_as_dim_region_id(self):
        """The method is testing getting data from dataset with region dimention by region id"""

        data_frame = knoema.get('IMFWEO2019Oct',
            region = 'AF',
            subject = ['Gross domestic product, constant prices (Percent change)']
        )

        self.assertEqual(['Country', 'Subject', 'Frequency'], data_frame.columns.names)
        self.assertEqual('Afghanistan', data_frame.columns.values[0][0])

        indx = data_frame.first_valid_index()
        sname = ('Afghanistan', 'Gross domestic product, constant prices (Percent change)', 'A')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 8.692)

    def test_getdata_with_columns(self):
        """The method is testing getting data from dataset with columns"""

        data_frame = knoema.get('pqgusj', Company = 'APPLE', columns='*', frequency='FQ')

        self.assertEqual(['Company', 'Indicator', 'Country', 'Frequency', 'Attribute'], data_frame.columns.names)
        self.assertEqual(3, data_frame.columns.size)
        self.assertEqual(('APPLE', 'Export', 'US', 'FQ', 'StatisticalDate'), data_frame.columns.values[1])

    def test_getdata_datelabels(self):
        """The method is testing getting data from dataset with region dimention by region id"""

        data_frame = knoema.get('pqgusj', company = 'APPLE', frequency='FQ')

        self.assertEqual(['Company', 'Indicator', 'Country', 'Frequency'], data_frame.columns.names)
        sname = ('APPLE', 'Export', 'US', 'FQ')

        # this is FQ dateLabel
        value = data_frame.at[pandas.to_datetime('2019-10-09'), sname]

        self.assertEqual(value, 0.531275885712192)

    def test_grouping_functionality(self):
        """The method is tesing grouping functionality"""

        generator = knoema.get('IMFWEO2017Oct', include_metadata = True, group_by='country', country='Albania;Italy;Japan', subject='ngdp', timerange='2005-2010')
        frames = []
        for frame in generator:
            frames.append(frame)

        self.assertEqual(len(frames), 3)
        self.assertIs(type(frames[0].data), pandas.core.frame.DataFrame)
        self.assertIs(type(frames[0].metadata), pandas.core.frame.DataFrame)

    def test_FQ_frequescy(self):
        """Testing FQ frequency"""

        data_frame = knoema.get('xmhdwqf', company='UBER', indicator='Fisqal Quarterly', frequency='FQ')
        self.assertIs(type(data_frame), pandas.core.frame.DataFrame)
        self.assertEqual(data_frame.shape[0], 26)
        self.assertEqual(data_frame.shape[1], 1)

    def test_aggregations(self):
        """Testing aggregations disaggregation"""

        frame = knoema.get('xmhdwqf', company='UBER', indicator='Daily', frequency='D', timerange = '2019M1-2021M4')
        self.assertEqual(frame.shape[0], 366)
        self.assertEqual(frame.shape[1], 1)

        generator = knoema.get('xmhdwqf', group_by = 'company', company='UBER', indicator='Daily', frequency='D', timerange = '2019M1-2021M4')
        for frame in generator:
            self.assertEqual(frame.data.shape[0], 366)
            self.assertEqual(frame.data.shape[1], 1)
            

        frame = knoema.get('xmhdwqf', company='UBER', indicator='Daily', frequency='Q', timerange = '2019M1-2021M4', transform='sum')
        self.assertEqual(frame.shape[0], 5)
        self.assertEqual(frame.shape[1], 1)

        generator = knoema.get('xmhdwqf', group_by='company', company='UBER', indicator='Daily', frequency='Q', timerange = '2019M1-2021M4', transform='sum')
        for frame in generator:
            self.assertEqual(frame.data.shape[0], 5)
            self.assertEqual(frame.data.shape[1], 1)

    def test_auto_aggregations_nodata(self):
        """Testing that auto aggregations returns no data"""

        frame = knoema.get('xmhdwqf', company='UBER', indicator='Daily', frequency='D',
                           timerange='2019M1-2021M4')
        self.assertEqual(frame.shape[0], 366)
        self.assertEqual(frame.shape[1], 1)

        frame = knoema.get('xmhdwqf', company='UBER', indicator='Daily', frequency='Q',
                           timerange='2019M1-2021M4')
        self.assertEqual(frame.shape[0], 0)
        self.assertEqual(frame.shape[1], 0)

    def test_auto_disaggregations_nodata(self):
        """Testing that auto disaggregations returns no data"""

        frame = knoema.get('xmhdwqf', company='UBER', indicator='Annual', frequency='A', timerange='2018-2020')
        self.assertEqual(frame.shape[0], 3)
        self.assertEqual(frame.shape[1], 1)

        frame = knoema.get('xmhdwqf', company='UBER', indicator='Annual', frequency='M', timerange='2018-2020')
        self.assertEqual(frame.shape[0], 0)
        self.assertEqual(frame.shape[1], 0)



import tempfile, os, shutil, string, random
import pandas as pd
from abc import ABC, abstractmethod
import csv

class FrameTransformerFactory(object):

    def __init__(self, frame):
        self._frame = frame

    def _get_column_types(self):
        date_columns = []
        value_columns = []
        dim_columns = []

        for col in self._frame.columns:
            is_numeric = True
            is_date = True

            try:
                pd.to_numeric(self._frame[col])
                value_columns.append(col)
            except ValueError:
                is_numeric = False

            if not is_numeric:
                try:
                    pd.to_datetime(self._frame[col])
                    date_columns.append(col)
                except ValueError:
                    is_date = False

            if not is_numeric and not is_date:
                dim_columns.append(col)

        return dim_columns, date_columns, value_columns

    def get_transformer(self):
        dim_columns, date_columns, value_columns = self._get_column_types()

        if len(date_columns) > 1 or (len(dim_columns) > 0 and len(value_columns) > 1):
            return FrameTransformerFlat(self._frame, dim_columns, date_columns, value_columns)
        
        return FrameTransformerRegular(self._frame, dim_columns, date_columns, value_columns)

class FrameTransformerBase(ABC):

    def __init__(self, frame, dim_columns, date_columns, value_columns):
        self._frame = frame
        self._dim_columns = dim_columns
        self._date_columns = date_columns
        self._value_columns = value_columns

        if frame is None:
            raise ValueError('Frame is not specified.')

        if frame.empty:
            raise ValueError('Frame cannot be empty.')

    @abstractmethod
    def prepare(self, dataset = None, dataset_name = None):
        return

class FrameTransformerFlat(FrameTransformerBase):

    def __init__(self, frame, dim_columns, date_columns, value_columns):
        super().__init__(frame, dim_columns, date_columns, value_columns)

    def _rows(self, frame):
        data_rows = []
        first_data_row = []

        for col in frame.columns:
            first_data_row.append(col)

        data_rows.append(first_data_row)
        nan_frame = frame.isnull()

        for i in range(len(frame.values)):
            frame_row = frame.values[i]
            nan_frame_row = nan_frame.values[i]

            row = []

            for j in range(len(frame_row)):
                item = frame_row[j]
                nan_item = nan_frame_row[j]
                row.append('' if nan_item else str(item))

            data_rows.append(row)

        return data_rows

    def prepare(self, file_wrapper, dataset = None, dataset_name = None):
        rows = self._rows(self._frame)

        file_name = (dataset_name if dataset_name != None else dataset) + '.csv'
        return file_wrapper.write_single_file(file_name, rows)

class FrameTransformerRegular(FrameTransformerBase):

    def __init__(self, frame, dim_columns, date_columns, value_columns):
        super().__init__(frame, dim_columns, date_columns, value_columns)

    def _get_axis_types(self):
        return [type(axes).__name__ for axes in self._frame.axes]

    def _change_columns_types(self, date_columns, value_columns):
        for date_column in date_columns:
            self._frame[date_column] = pd.to_datetime(self._frame[date_column])

        for value_column in value_columns:
            self._frame[value_column] = pd.to_numeric(self._frame[value_column])

    def _prepare_frame(self):
        ready_to_upload_frame = None

        axes_types = self._get_axis_types()

        self._change_columns_types(self._date_columns, self._value_columns)

        if 'DatetimeIndex' not in axes_types:
            if len(self._date_columns) == 0:
                raise ValueError('The frame has no column with dates.')

            if len(self._date_columns) == 1:
                self._frame.set_index(self._date_columns[0], drop = True, append = False, inplace = True)

            if len(self._value_columns) > 1:
                raise ValueError('The frame has few columns with dates.')

        if 'MultiIndex' in axes_types:
            ready_to_upload_frame = self._frame
        else:
            if len(self._dim_columns) == 0:
                if len(self._value_columns) == 0:
                    raise ValueError('The frame doesn\'t have column with values.')

                ready_to_upload_frame = self._frame
            else:
                if len(self._value_columns) == 0:
                    raise ValueError('The frame doesn\'t have column with values.')

                if len(self._value_columns) == 1:
                    ready_to_upload_frame = self._frame.pivot_table(
                        index = self._date_columns[0],
                        columns = self._dim_columns if len(self._dim_columns) > 0 else None,
                        values = self._value_columns[0]
                    )
                else:
                    raise ValueError('The frame has more than one column with values.')
                
        return ready_to_upload_frame

    def _parse_dimension_list(self, full_dim_list):
        freq_index = -1
        scale_index = -1
        unit_index = -1

        dimensions = []

        if len(full_dim_list) == 1 and full_dim_list[0] == None:
            full_dim_list = ['Dimension1']

        for i in range(len(full_dim_list)):
            dim = full_dim_list[i]
            if dim == 'Frequency':
                freq_index = i
                continue

            if dim == 'Scale':
                scale_index = i
                continue

            if dim == 'Unit':
                unit_index = i
                continue

            dimensions.append(full_dim_list[i])

        return dimensions, freq_index, scale_index, unit_index

    def _dataset_sheet(self, dataset_id, dataset_name, dimensions):
        rows = []
        rows.append(['Name', 'Value'])
        if dataset_id != None:
            rows.append(['Dataset', dataset_id])
        if dataset_id != None or dataset_name != None:
            rows.append(['Dataset name', dataset_name if dataset_name != None else dataset_id])
        rows.append(['Dimensions', ';'.join(dimensions)])
        rows.append(['Data', 'Data'])
        return rows

    def _get_unic_id(self, name, map):
        if name in map:
            return map[name]

        parts_raw = name.split(' ')
        parts = []
        for part in parts_raw:
            new_part = ''.join(p for p in part if p.isalnum())
            if new_part == '':
                continue
                    
            parts.append(new_part.upper())

        if len(parts) == 1:
            return parts[0].upper()[:8]

        if len(parts) <= 3:
            res = ''
            for part in parts:
                res += part.upper()[:3]
            return res

        res = ''
        for part in parts:
            res += part.upper()[0]

        counter = 1
        res_with_counter = res
        while True:
            if res_with_counter in map.values():
                res_with_counter = res + '_' + str(counter)
                counter += 1
            else:
                res = res_with_counter
                break
                
        return res

    def _dimension_sheets(self, dimensions, series_names):
        dimensions_rows = {}
        dimensions_map = {}
        
        for dim_ind in range(len(dimensions)):
            dim = dimensions[dim_ind]
            dim_rows = []
            dim_map = {}
            first_row = ['Name', 'Code']

            dim_rows.append(first_row)

            for name in series_names:
                part = name if len(dimensions) == 1 else name[dim_ind]
                id = self._get_unic_id(part, dim_map)
                row = [part, id]

                if part in dim_map:
                    continue

                dim_map[part] = id

                dim_rows.append(row)

            dimensions_rows[dim] = dim_rows
            dimensions_map[dim] = dim_map

        return dimensions_rows, dimensions_map

    def _freq_fetch(self, pandas_freq):
        parts = pandas_freq.split('-')
        freq = parts[0]

        # https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        if freq in ['B', 'C', 'D']: return 'D'
        if freq in ['W']: return 'W'
        if freq in ['M', 'SM', 'BM', 'CBM', 'MS', 'BMS', 'CBMS']: return 'M'
        if freq in ['Q', 'BQ', 'QS', 'BQS']: return 'Q'
        if freq in ['A', 'Y', 'BA', 'BY', 'AS', 'YS', 'BAS', 'BYS']: return 'A'

        return None

    def _data_sheet(self, frame, dimensions, dimensions_map, freq_index, scale_index, unit_index, series_names):
        data_rows = []
        first_data_row = []
        for d_n in dimensions:
            first_data_row.append(d_n)

        if scale_index >= 0:
            first_data_row.append('Scale')

        if unit_index >= 0:
            first_data_row.append('Unit')

        first_data_row.append('Frequency')

        first_data_row.extend(['Date', 'Value'])
        data_rows.append(first_data_row)

        for name in series_names:
            series = frame[name]
            nan_series = series.isnull()
            for ind in range(len(frame.index)):
                row = []

                if len(dimensions) == 1:
                    map = dimensions_map[dimensions[0]]
                    row.append(map[name])

                    freq = None
                    if frame.axes[0].freq == None:
                        freq = self._freq_fetch(pd.infer_freq(frame.axes[0]))
                    else:
                        freq = self._freq_fetch(self._frame.axes[0].freq.name)

                    if freq == None:
                        raise ValueError('Wrong frequency.')

                    row.append(freq)
                else:
                    for part_ind in range(len(name)):
                        if scale_index >= 0 and part_ind == scale_index:
                            continue

                        if unit_index >= 0 and part_ind == unit_index:
                            continue

                        if freq_index >= 0 and part_ind == freq_index:
                            continue
                            
                        part = name[part_ind]

                        dim_name = frame.columns.names[part_ind]
                        if dim_name in dimensions_map:
                            map = dimensions_map[dim_name]
                            part = map[part]

                        row.append(part)

                    if scale_index >= 0:
                        row.append(name[scale_index])

                    if unit_index >= 0:
                        row.append(name[unit_index])

                    if freq_index >= 0:
                        row.append(name[freq_index])

                    if freq_index == -1:
                        freq = None
                        if frame.axes[0].freq == None:
                            freq = self._freq_fetch(pd.infer_freq(frame.axes[0]))
                        else:
                            freq = self._freq_fetch(self._frame.axes[0].freq.name)

                        if freq == None:
                            raise ValueError('Wrong frequency.')

                        row.append(freq)

                row.append(frame.index[ind])
                row.append(series[ind] if not nan_series[ind] else '')

                data_rows.append(row)

        return data_rows

    def prepare(self, file_wrapper, dataset = None, dataset_name = None):
        frame = self._prepare_frame()

        dimensions, freq_index, scale_index, unit_index = self._parse_dimension_list(frame.columns.names)

        dataset_rows = self._dataset_sheet(dataset, dataset_name, dimensions)

        series_names = frame.columns.values
        dimensions_rows, dimensions_map = self._dimension_sheets(dimensions, series_names)

        data_rows = self._data_sheet(frame, dimensions, dimensions_map, freq_index, scale_index, unit_index, series_names)

        file_wrapper.add_to_archive('Dataset.csv', dataset_rows)
        for dim in dimensions_rows:
            file_wrapper.add_to_archive('{}.csv'.format(dim), dimensions_rows[dim])
        file_wrapper.add_to_archive('Data.csv', data_rows)

        return file_wrapper.get_archive()

class FileLayerWrapper(object):

    def __init__(self):
        self._tmp_dir = None
        self._tmp_file = None

    def write_single_file(self, name, rows):
        parent_folder = tempfile.gettempdir()
        self._write_file(parent_folder, name, rows)

        self._tmp_file = os.path.join(parent_folder, name)
        return self._tmp_file

    def add_to_archive(self, name, rows):
        if self._tmp_dir == None:
            self._create_tmp_dir()

        self._write_file(self._tmp_dir, name, rows)

    def get_archive(self):

        shutil.make_archive(self._tmp_dir, 'zip', self._tmp_dir)
        self._tmp_file = self._tmp_dir + '.zip'
        return self._tmp_file

    def _create_tmp_dir(self):
        parent_folder = tempfile.gettempdir()
        tmp_dir = os.path.join(parent_folder, self._get_tmp_dir_name())

        os.makedirs(tmp_dir, exist_ok=True)

        self._tmp_dir = tmp_dir

    def _get_tmp_dir_name(self, len = 8):
        letters = string.ascii_lowercase
        return ''.join(random.choice(letters) for i in range(len))

    def _write_file(self, path, name, rows):
        with open(os.path.join(path, name), 'w', newline = '') as file:
            writer = csv.writer(file)
            writer.writerows(rows)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._tmp_dir != None and os.path.exists(self._tmp_dir):
            shutil.rmtree(self._tmp_dir)

        if self._tmp_file != None and os.path.isfile(self._tmp_file):
            os.remove(self._tmp_file)




"""This is test module for knoema client with test credentials"""

import unittest
import knoema
import os

class TestKnoemaClient(unittest.TestCase):
    """This is class with knoema client unit tests with test credentials"""

    base_host = 'knoema.com'

    def setUp(self):
        apicfg = knoema.ApiConfig()
        apicfg.host = os.environ['BASE_HOST'] if 'BASE_HOST' in os.environ else self.base_host
        apicfg.app_id = os.environ['KNOEMA_APP_ID'] if 'KNOEMA_APP_ID' in os.environ else ''
        apicfg.app_secret = os.environ['KNOEMA_APP_SECRET'] if 'KNOEMA_APP_SECRET' in os.environ else ''

    def test_getdata_by_TransformationDataReader_timemembers(self):
        """The method is testing getting single series by dimension member ids"""

        data_frame = knoema.get('IMFWEO2021Apr', Country='614', Subject='BCA', timemembers='1980;2002;2023')
        self.assertEqual(data_frame.shape[0], 3)
        self.assertEqual(data_frame.shape[1], 1)

        self.assertEqual(['Country', 'Subject', 'Frequency'], data_frame.columns.names)

        indx = data_frame.first_valid_index()
        sname = ('Angola', 'Current account balance (U.S. dollars)', 'A')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 0.07)

        indx = data_frame.index[2]
        value = data_frame.at[indx, sname]
        self.assertEqual(value, -0.154)

    def test_getdata_by_TransformationDataReader_timesince(self):
        """The method is testing getting single series by dimension member ids"""

        data_frame = knoema.get('IMFWEO2021Apr', Country='614', Subject='BCA', timesince='2013')
        self.assertEqual(data_frame.shape[0], 14)
        self.assertEqual(data_frame.shape[1], 1)

        self.assertEqual(['Country', 'Subject', 'Frequency'], data_frame.columns.names)

        indx = data_frame.first_valid_index()
        sname = ('Angola', 'Current account balance (U.S. dollars)', 'A')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 8.348)

        indx = data_frame.index[13]
        value = data_frame.at[indx, sname]
        self.assertEqual(value, -0.457)

    def test_getdata_by_TransformationDataReader_timelast1(self):
        """The method is testing getting multiple series by dimension member ids and time range"""

        data_frame = knoema.get('IMFDOT2017', **{'Country': 'Algeria;Angola', 'Indicator': 'TXG_FOB_USD', 'Counterpart Country': '622', 'frequency': 'A;Q', 'timelast': '1'})

        self.assertEqual(data_frame.shape[0], 4)
        self.assertEqual(data_frame.shape[1], 4)



"""This module contains Api configuration class"""

import os

class ApiConfig(object):
    """
    This class configures knoema api.

    The class contains fields:

    host -- the host where kneoma is going to connect

    app_id -- application id that will have access to knoema.
    Application should be created by knoema user or administrator

    app_secret -- code that can be done after application will be created.
    Should be set up together with app_id
    """

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(ApiConfig, cls).__new__(cls)
            cls.instance.host = os.environ['KNOEMA_HOST'] if 'KNOEMA_HOST' in os.environ else 'knoema.com'
            cls.instance.app_id = None
            cls.instance.app_secret = None
        return cls.instance

    def __init__(self):
        self.host = self.instance.host
        self.app_id = self.instance.app_id
        self.app_secret = self.instance.app_secret



import pytest
import knoema
import pandas


@pytest.mark.parametrize('dataset_id,dimension_id', [
    ('IMFWEO2020Apr', 'country'),
    ('INTLRRC2017', 'country'),
    ('bjxchy', 'loan-type')])
def test_dimension_metadata(get_client, dataset_id, dimension_id):
    res = knoema.dimension(dataset_id, dimension_id)

    assert res.datasetId == dataset_id
    assert res.id == dimension_id
    assert type(res.members) == pandas.core.frame.DataFrame

def test_dimension_metadata_not_all_fields_in_items(get_client):
    dataset_id = 'kaziajg'
    dimension_name = 'Location'
    res = knoema.dimension(dataset_id, dimension_name)

    assert res.datasetId == dataset_id
    assert res.name == dimension_name
    assert type(res.members) == pandas.core.frame.DataFrame
    assert len(res.members.axes[1]) == 12



class Dimension:

    def __init__(self):
        self.key = None
        self.id = None
        self.name = None
        self.isGeo = None
        self.datasetId = None
        self.fields = []
        self.members = None


class Field:

    def __init__(self, field_info):
        self.key = field_info['key']
        self.name = field_info['name']
        self.displayName = field_info['displayName']
        self.type = field_info['type']
        self.locale = field_info['locale']
        self.baseKey = field_info['baseKey']
        self.isSystemField = field_info['isSystemField']



import knoema

def test_calculations_in_dimention(get_client):
    data = knoema.get('IMFWEO2020Apr', **{
    'Country': 'Armenia',
    'Subject': '@SUM;NGDP_RPCH;PCPIPCH'})

    assert data.columns.values[0] == ('Armenia', 'Sum of 2', 'A')
    assert data.shape == (29, 1)


"""This module contains metadata definitions for Knoema API"""

import json
from datetime import datetime

def is_equal_strings_ignore_case(first, second):
    """The function compares strings ignoring case"""
    if first and second:
        return first.upper() == second.upper()
    else:
        return not (first or second)


def parse_date(date_str):
    return datetime.strptime(date_str.split('.')[0].strip('Z'), '%Y-%m-%dT%H:%M:%S') if date_str != None else None


class DatasetMetadata:

    def __init__(self, data):
        self.data = data

        settings = {}
        if 'settings' in self.data:
            try:
                settings = json.loads(self.data['settings'])
            except:
                pass

            if 'LastUpdate' in settings:
                data['lastUpdate'] = settings['LastUpdate']['End']
            
            if 'LastSuccessfulUpdate' in settings:
                data['lastSuccessfulUpdate'] = settings['LastSuccessfulUpdate']['End']

            if 'Health' in settings:
                data['health'] = settings['Health']

            if 'NextRun' in settings:
                data['nextRun'] = settings['NextRun']

            if 'AverageUpdateInterval' in settings:
                data['averageUpdateInterval'] = settings['AverageUpdateInterval']

            if 'ProviderUpdateLag' in settings:
                data['providerUpdateLag'] = settings['ProviderUpdateLag']

        fields_with_date = [
            'publicationDate',
            'nextReleaseDate',
            'expectedUpdateDate',
            'lastUpdatedOn',
            'lastUpdate',
            'lastSuccessfulUpdate',
            'nextRun',
        ]
        for field in fields_with_date:
            if field in self.data:
                self.data[field] = parse_date(self.data[field])

        fields_to_delete = [
            'updatePriority',
            'hasShortCut',
            'isShortcut',
            'shareToCommunitiesAllowed',
            'accessedOn',
            'metadataAccess',
            'settings',
        ]
        for field in fields_to_delete:
            if field in self.data:
                del self.data[field] 


class DimensionMember(object):
    """The class contains dimension member information"""

    def __init__(self, data):
        self.key = data['key']
        self.name = data['name']
        self.level = data['level']
        self.hasdata = data['hasData']
        self.fields = data['fields']


class DimensionModel(object):
    """The class contains dimension description"""

    def __init__(self, data):
        self.key = data['key']
        self.id = data['id']
        self.name = data['name']
        self.is_geo = data['isGeo'] if 'isGeo' in data else False


class Dimension(DimensionModel):
    """The class contains dimension description and dimension items"""

    def __init__(self, data):
        super().__init__(data)
        self.fields = data['fields']
        self.items = [DimensionMember(item) for item in data['items']]

        # fill maps
        self.key_map = {}
        self.id_map = {}
        self.name_map = {}
        self.region_map = {}
        self.ticker_map = {}
        for item in self.items:
            self.key_map[item.key] = item
            self.name_map[item.name.upper()] = item
            if 'id' in item.fields:
                self.id_map[item.fields['id'].upper()] = item
            if 'regionid' in item.fields and item.fields['regionid'] is not None:
                self.region_map[item.fields['regionid'].upper()] = item
            if 'ticker' in item.fields and item.fields['ticker'] is not None:
                self.ticker_map[item.fields['ticker'].upper()] = item

        
    def find_member_by_key(self, member_key):
        """The method searches member of dimension by given member key"""
        return self.key_map.get(member_key)

    def find_member_by_id(self, member_id):
        """The method searches member of dimension by given member id"""
        return self.id_map.get(member_id.upper())

    def find_member_by_name(self, member_name):
        """The method searches member of dimension by given member name"""
        return self.name_map.get(member_name.upper())

    def find_member_by_regionid(self, member_name):
        """The method searches member of dimension by given region id"""
        return self.region_map.get(member_name.upper())

    def find_member_by_ticker(self, member_name):
        return self.ticker_map.get(member_name.upper())


class DateRange:
    """The class contains information about dataset's data range"""

    def __init__(self, data):
        self.start_date = parse_date(data['startDate'])
        self.end_date = parse_date(data['endDate'])
        self.frequencies = data['frequencies']


class TimeSeriesAttribute(object):
    """This class contains information about dataset' s timeseries attributes"""
    def __init__(self, data):
        self.name = data['name']
        self.type = data['type']
        self.allowed_values = data['allowedValues']

class Dataset(object):
    """The class contains dataset description"""

    def __init__(self, data):
        """The method loading data from json to class fields"""

        if 'id' not in data:
            raise ValueError(data)

        self.id = data['id']
        self.type = data['type']
        self.is_remote = data['isRemote'] if 'isRemote' in data else False
        self.dimensions = [DimensionModel(dim) for dim in data['dimensions']]
        self.timeseries_attributes = [TimeSeriesAttribute(attr) for attr in data['timeseriesAttributes']] if 'timeseriesAttributes' in data else []
        self.has_time = self.type == 'Regular' or any(x for x in data['columns'] if x['type'] == 'Date')
            
    def find_dimension_by_name(self, dim_name):
        """the method searching dimension with a given name"""

        for dim in self.dimensions:
            if is_equal_strings_ignore_case(dim.name, dim_name):
                return dim

            if dim.is_geo and is_equal_strings_ignore_case('region', dim_name):
                return dim
        return None

    def find_dimension_by_id(self, dim_id):
        """the method searching dimension with a given id"""

        for dim in self.dimensions:
            if is_equal_strings_ignore_case(dim.id, dim_id):
                return dim

            if dim.is_geo and is_equal_strings_ignore_case('region', dim_id):
                return dim

            if is_equal_strings_ignore_case('ticker', dim_id):
                return dim

        return None

class PivotItemMetadata(object):
#SimpleDimensionMember
    """The class contains metadata fields for pivot request item"""
    def __init__(self, key, name, parent=None, fields=None):
        self.key = key
        self.name = name
        self.parent = parent
        self.fields = fields

class PivotItem(object):
    """The class contains pivot request item"""

    def __init__(self, dimensionid=None, members=None, metadataFields=None, dimensionFields=None, aggregations=None):
        self.dimensionid = dimensionid
        self.members = members
        if aggregations != None:
            self.aggregation = aggregations
        if metadataFields:
            self.metadataFields = [PivotItemMetadata(metadata['key'], metadata['name'],
                metadata['parent'], metadata['fields']) for metadata in metadataFields]
        else:
            self.metadataFields = None
        self.fields = dimensionFields


class PivotTimeItem(PivotItem):
    """The class contains pivot request item"""

    def __init__(self, dimensionid=None, members=None, uimode=None):
        super().__init__(dimensionid, members)
        self.uimode = uimode


class PivotRequest(object):
    """The class contains pivot request"""

    def __init__(self, dataset):
        self.dataset = dataset
        self.header = []
        self.stub = []
        self.filter = []
        self.frequencies = []
        self.transform = None
        self.columns = None

    def _get_item_array(self, items):
        arr = []
        for item in items:
            itemvalues = {
                'DimensionId': item.dimensionid,
                'Members': item.members
            }
            if hasattr(item, 'aggregation'):
                itemvalues['Aggregation'] = item.aggregation
            if isinstance(item, PivotTimeItem):
                itemvalues['UiMode'] = item.uimode
            arr.append(itemvalues)
        return arr

    def save_to_json(self):
        """The method saves data to json from object"""
        requestvalues = {
            'Dataset': self.dataset,
            'Header' : self._get_item_array(self.header),
            'Filter' : self._get_item_array(self.filter),
            'Stub' : self._get_item_array(self.stub),
            'Frequencies': self.frequencies
        }
        if self.transform is not None:
            requestvalues['Transform'] = self.transform
        if self.columns is not None:
            requestvalues['DetailColumns'] = self.columns
            
        return json.dumps(requestvalues)

class DataAPIRequest(object):
    """The class contains API Data 2.0 request"""

    def __init__(self, filters):
        self.filters = filters

    def save_to_json(self):
        """The method saves data to json from object"""
        return json.dumps(self.filters)

class DetailsResponse(object):
    def __init__(self, data):
        self.columns = data['columns']
        self.tuples = data['data']

class PivotResponse(object):
    """The class contains pivot response"""

    def __init__(self, data):

        self.dataset = data['dataset']

        self.header = []
        for item in data['header']:
            self.header.append(self._construct_dimension(item))

        self.stub = []
        for item in data['stub']:
            self.stub.append(self._construct_dimension(item))

        self.filter = []
        for item in data['filter']:
            self.filter.append(self._construct_dimension(item))

        self.tuples = data['data']
        self.descriptor = data['descriptor'] if 'descriptor' in data else None

    def _construct_dimension(self, item):
        return PivotItem(item['dimensionId'], item['members'], item['metadataFields'], item['dimensionFields'] if 'dimensionFields' in item else None)

class RawDataResponse(object):
    """The class contains raw data response"""

    def __init__(self, data):

        self.continuation_token = data['continuationToken']
        self.series = data['data']
        self.descriptor = data['descriptor'] if 'descriptor' in data else None
        self.dimensionFields = data['dimensionFields'] if 'dimensionFields' in data else None

def detect_data_response(data):
    return DetailsResponse(data) if 'columns' in data \
        else PivotResponse(data) if 'keys' in data \
        else RawDataResponse(data) if 'continuationToken' in data else None

class MnemonicsResponseList(object):

    def __init__(self,data):
        self.items = []
        for item in data:
            self.items.append(MnemonicsResponse(item))   

class MnemonicsResponse(object):

    def __init__(self,data):
        self.mnemonics = data['mnemonics']
        self.pivot = PivotResponse(data['pivot']) if 'pivot' in data else None

class FileProperties(object):
    """The class contains response from upload post request"""

    def __init__(self, data):
        self.size = data['Size'] if 'Size' in data else None
        self.name = data['Name'] if 'Name' in data else None
        self.location = data['Location'] if 'Location' in data else None
        self.type = data['Type'] if 'Type' in data else None


class UploadPostResponse(object):
    """The class contains response from upload post request"""

    def __init__(self, data):
        self.successful = data['Successful'] if 'Successful' in data else False
        self.error = data['Error'] if 'Error' in data else None
        self.properties = FileProperties(data['Properties'])


class UploadDatasetDetails(object):
    """The class contains dataset details from verify result response"""

    def __init__(self, data):
        self.dataset_id = data['DatasetId']
        self.dataset_name = data['DatasetName']
        self.source = data['Source']
        self.description = data['Description']
        self.dataset_ref = data['DatasetRef']
        self.publication_date = data['PublicationDate'] if 'PublicationDate' in data else None
        self.accessed_on = data['AccessedOn'] if 'AccessedOn' in data else None


class UploadVerifyResponse(object):
    """The class contains response from upload post request"""

    def __init__(self, data):
        self.successful = data['Successful'] if 'Successful' in data else False
        self.upload_format_type = data['UploadFormatType'] if 'UploadFormatType' in data else None
        self.errors = data['ErrorList'] if 'ErrorList' in data else None
        self.columns = data['Columns'] if 'Columns' in data else None
        self.flat_ds_update_options = data['FlatDSUpdateOptions'] if 'FlatDSUpdateOptions' in data else None
        self.metadata_details = UploadDatasetDetails(data['MetadataDetails']) if 'MetadataDetails' in data and data['MetadataDetails'] is not None else None


class DatasetUpload(object):
    """The class contains request for UploadSubmit"""

    def __init__(self, verify_result, upload_result, dataset = None, public = False, name = None):
        self.dataset = dataset

        self.upload_format_type = verify_result.upload_format_type
        self.columns = verify_result.columns
        self.file_property = upload_result.properties
        self.flat_ds_update_options = verify_result.flat_ds_update_options

        dataset_details = verify_result.metadata_details
        self.name = dataset_details.dataset_name if dataset_details and dataset_details.dataset_name else None
        if self.name is None and dataset is None:
            self.name = name if name != None else 'New dataset'
        self.description = dataset_details.description if dataset_details else None
        self.source = dataset_details.source if dataset_details else None
        self.publication_date = dataset_details.publication_date if dataset_details else None
        self.accessed_on = dataset_details.accessed_on if dataset_details else None
        self.dataset_ref = dataset_details.dataset_ref if dataset_details else None
        self.public = public

    def save_to_json(self):
        """The method saves DatasetUpload to json from object"""
        requestvalues = {
            'DatasetId': self.dataset,
            'Name': self.name,
            'Description': self.description,
            'Source': self.source,
            'PubDate': self.publication_date,
            'AccessedOn': self.accessed_on,
            'Url': self.dataset_ref,
            'UploadFormatType': self.upload_format_type,
            'Columns': self.columns,
            'FileProperty': self.file_property.__dict__,
            'FlatDSUpdateOptions': self.flat_ds_update_options,
            'Public': self.public
        }
        return json.dumps(requestvalues)


class DatasetUploadResponse(object):
    """The class contains response for UploadSubmit"""

    def __init__(self, data):
        self.submit_id = data['Id'] if 'Id' in data else None
        self.dataset = data['DatasetId'] if 'DatasetId' in data else None
        self.status = data['Status'] if 'Status' in data else 'failed'
        self.errors = data['Errors'] if 'Errors' in data else None


class DatasetUploadStatusResponse(object):
    """The class contains response for UploadSubmit"""

    def __init__(self, data):
        self.submit_id = data['id'] if 'id' in data else None
        self.dataset = data['datasetId'] if 'datasetId' in data else None
        self.status = data['status'] if 'status' in data else 'failed'
        self.errors = data['errors'] if 'errors' in data else None


class DatasetVerifyRequest(object):
    """The class contains dataset verification request"""

    def __init__(self, dataset, publication_date, source, refernce_url):
        self.dataset = dataset
        self.publication_date = publication_date
        self.source = source
        self.refernce_url = refernce_url

    def save_to_json(self):
        """The method saves data to json from object"""

        requestvalues = {
            'id': self.dataset,
            'publicationDate': self.publication_date.strftime('%Y-%m-%d'),
            'source': self.source,
            'refUrl': self.refernce_url,
        }        
        return json.dumps(requestvalues)


class DatasetVerifyResponse(object):
    """The class contains response from dataset verification request"""

    def __init__(self, data):
        self.status = data['status']
        self.errors = data['errors'] if 'errors' in data else None

class DataFrame(object):

    def __init__(self):
        self.id = None
        self.data = None
        self.metadata = None



from setuptools import setup

def readme():
    with open('README.rst', encoding="utf8") as readme_file:
        return readme_file.read()

setup(
  name = 'knoema',
  packages = ['knoema'],
  version = '1.3.0',
  description = "Official Python package for Knoema's API",
  long_description=readme(),
  author = 'Knoema',
  author_email = 'info@knoema.com',
  license='MIT',
  url = 'https://github.com/Knoema/knoema-python-driver',
  keywords = ['API', 'knoema'],
  classifiers = ['Development Status :: 5 - Production/Stable', 'Programming Language :: Python :: 3 :: Only'],
  install_requires=['pandas==2.2.2', 'pytest-shutil==1.7.0']
)


"""This module contains metadata definitions for Knoema API for semantic atlas"""

from knoema.data_reader import TransformationDataReader, PivotResponseReader

class Company(object):
    """"The class contains data related to a company like name and groups of indicators"""

    def __init__(self, company, client):
        self.name = company.name

        self.groups = []

        for group in company.groups:
            self.groups.append(CompanyIndicatorsGroup(group, client))

    def get_indicator(self, name, group = None):
        for grp in self.groups:
            if group != None and grp.name != group:
                continue

            for indicator in grp.indicators:
                if indicator.name == name:
                    return indicator
        
        return None


class CompanyIndicatorsGroup(object):
    """"The class contains information about named group of indicator"""

    def __init__(self, group, client):
        self.name = group.name
        
        self.indicators = []
        for indicator in group.indicators:
            self.indicators.append(CompanyIndicator(indicator, client))


class CompanyIndicator(object):
    """The class contains information about company indicator"""

    def __init__(self, indicator, client):
        self._client = client

        self._id = indicator._id
        self._full_id = indicator._full_id
        self.name = indicator.name
        self.count = indicator.count

    def get(self, transform = None):
        ind_info = self._client.get_indicator_info(self._full_id)
        if len(ind_info['groups']) < 1:
            return None

        first_group = ind_info['groups'][0]

        dataset = first_group['id']
        ds = self._client.get_dataset(dataset) if dataset else None

        desc = first_group['batchDesctiptor']
        pivot = self._client.get_data_by_json(desc)

        tr_reader =  TransformationDataReader(self._client, None, transform, None)
        tr_reader.dataset = ds
        pivot_reader = PivotResponseReader(tr_reader, pivot)
        frame = pivot_reader.get_pandasframe()

        return frame


class CompanyInt(object):

    def __init__(self, data):
        self.name = data['title']

        self.groups = []
        for group in data['groupHierarchies']:
            self.groups.append(CompanyIndicatorsGroupInt(group))
        

class CompanyIndicatorsGroupInt(object):

    def __init__(self, data):
        self.name = data['topParent']['name']

        hierarchy = data['hierarchy']
        item_data = data['itemData']
        self.indicators = []
        for _, indicators_list in hierarchy.items():
            for ind in indicators_list:
                indicator = {
                    'id': ind['id'],
                    'name': ind['name']
                }

                ind_key = str(ind['key'])
                if ind_key in item_data:
                    indicator['full_id'] = item_data[ind_key]['conceptsQuery']
                    indicator['count'] = item_data[ind_key]['count']

                self.indicators.append(CompanyIndicatorInt(indicator))


class CompanyIndicatorInt(object):
    
    def __init__(self, data):
        self._id = data['id']
        self._full_id = data['full_id']
        self.name = data['name']
        self.count = data['count']



"""This is test module for knoema client"""
"""Works only with special credentials"""

import unittest
import knoema
import pandas
import os

class TestKnoemaClient(unittest.TestCase):
    """This is class with knoema client unit tests"""

    base_host = 'knoema.com'

    def setUp(self):
        apicfg = knoema.ApiConfig()
        apicfg.host = os.environ['BASE_HOST'] if 'BASE_HOST' in os.environ else self.base_host
        apicfg.app_id = os.environ['KNOEMA_APP_ID'] if 'KNOEMA_APP_ID' in os.environ else ''
        apicfg.app_secret = os.environ['KNOEMA_APP_SECRET'] if 'KNOEMA_APP_SECRET' in os.environ else ''

    def test_too_long_get_query_string(self):
        """The method is testing issue with loo long get query string"""

        for frame in knoema.get('US500COMPFINST2017Oct', include_metadata=True, indicator='Total Assets',
                                frequency='A', group_by='company'):
            company = frame.id
            data_frame = frame.data
            metadata_frame = frame.metadata

            self.assertIsNotNone(company, None)
            self.assertIs(type(data_frame), pandas.core.frame.DataFrame)
            self.assertIs(type(metadata_frame), pandas.core.frame.DataFrame)

            break

    def test_too_long_data_url(self):
        """The method is testing issue with loo long get query string"""
        subject = ['Gross domestic product, constant prices (Percent change)',
                   'Gross domestic product, constant prices (Percent change (market exchange rates))',
                   'Gross domestic product, current prices (U.S. dollars)',
                   'Rubber, No.1 Rubber Smoked Sheet, FOB Maylaysian/Singapore, US cents per pound (U.S. cents)',
                   'Gross domestic product, current prices (Purchasing power parity; international dollars)',
                   'Gross national savings (Percent of GDP)',
                   'Total investment (Percent of GDP)',
                   'Inflation, average consumer prices (Percent change)',
                   'Inflation, end of period consumer prices (Percent change)',
                   'Trade volume of goods and services (Percent change)',
                   'Volume of imports of goods and services (Percent change)',
                   'Volume of Imports of goods (Percent change)',
                   'Volume of exports of goods and services (Percent change)',
                   'Volume of exports of goods (Percent change)',
                   'Current account balance (U.S. dollars)',
                   'Commodity Price Index includes both Fuel and Non-Fuel Price Indices (Index, 2005=100)',
                   'Commodity Non-Fuel Price Index includes Food and Beverages and Industrial Inputs Price Indices (Index, 2005=100)',
                   'Commodity Industrial Inputs Price Index includes Agricultural Raw Materials and Metals Price Indices (Index, 2005=100)',
                   'Coal, Australian thermal coal, 1200- btu/pound, less than 1% sulfur, 14% ash, FOB Newcastle/Port Kembla, US$ per metric tonne (U.S. dollars)',
                   'Coal, South African export price, US$ per metric tonne (U.S. dollars)',
                   'Commodity Coal Price Index includes Australian and South African Coal (Index, 2005=100)',
                   'Commodity Fuel (energy) Index includes Crude oil (petroleum), Natural Gas, and Coal Price Indices (Index, 2005=100)',
                   'Commodity Natural Gas Price Index includes European, Japanese, and American Natural Gas Price Indices (Index, 2005=100)',
                   'Crude Oil (petroleum),  Dated Brent, light blend 38 API, fob U.K., US$ per barrel (U.S. dollars)',
                   'Crude Oil (petroleum), Price index simple average of three spot prices (APSP); Dated Brent, West Texas Intermediate, and the Dubai Fateh (Index, 2005=100)',
                   'Coffee, Robusta, International Coffee Organization New York cash price, ex-dock New York, US cents per pound (U.S. cents)',
                   'Commodity Beverage Price Index includes Coffee, Tea, and Cocoa (Index, 2005=100)',
                   'Commodity Cereals Price Index includes Wheat, Maize (Corn), Rice, and Barley (Index, 2005=100)',
                   'Commodity Coffee Price Index includes Other Mild Arabicas and Robusta (Index, 2005=100)']
        separator = ";;"
        frame = knoema.get("IMFWEO2021Apr", country="World", separator=separator, subject=separator.join(subject))
        self.assertEqual(frame.shape[1], len(subject))

    def test_ticker_endpoint(self):
        """Testing ticker endpoint"""

        ticker = knoema.ticker('DDD')

        self.assertEqual(ticker.name, '3D Systems Corporation')
        self.assertEqual(len(ticker.groups), 4)
        self.assertEqual(ticker.groups[0].name, 'Credit Risk')
        self.assertEqual(len(ticker.groups[0].indicators), 6)
        self.assertEqual(ticker.groups[0].indicators[0].name, 'Bankruptcy Base FRISK Probability')
        self.assertIs(type(ticker.groups[0].indicators[0].get()), pandas.core.frame.DataFrame)

        indicator = ticker.get_indicator('Bankruptcy Base FRISK Probability')
        self.assertEqual(indicator.name, 'Bankruptcy Base FRISK Probability')
        self.assertIs(type(indicator.get()), pandas.core.frame.DataFrame)

    def test_streaming_more_than_1000(self):
        """The method is testing getting multiple series by dimension member ids and time range"""
        data_frame = knoema.get('nama_r_e3gdp', **{'Measure': 'Euro per inhabitant'})
        self.assertEqual(data_frame.shape[0], 12)
        self.assertEqual(data_frame.shape[1], 1738)

    def test_search_wrapper_search_for_timeseries(self):
        """The method is testing search wrapper to search for timeseries"""

        search_results = knoema.search('Italy inflation')
        search_results_1 = knoema.search('UAE oil production')

        self.assertTrue(len(search_results.series) > 0)
        self.assertTrue(len(search_results_1.series) > 0)

        first_series = search_results.series[0].get()
        first_series_1 = search_results_1.series[0].get()

        self.assertIs(type(first_series), pandas.core.frame.DataFrame)
        self.assertIs(type(first_series_1), pandas.core.frame.DataFrame)

    def test_noaggregation(self):
        data = knoema.get(**{"dataset" : "IMFWEO2020Oct", "country": "United States", "subject": "Population (Persons)", "frequency" : "Q" , "transform": 'NOAGG'})
        self.assertEqual(data.shape[1], 0)

    def test_include_metadata_true(self):
        """The method is testing getting multiple series with data and metadata"""

        data, metadata = knoema.get('IMFWEO2017Apr', True, country=['914', '512'], subject='lp')
        self.assertEqual(data.shape[0], 43)
        self.assertEqual(data.shape[1], 2)
        self.assertEqual(['Country', 'Subject', 'Frequency'], data.columns.names)
        self.assertEqual(metadata.shape[0], 7)
        self.assertEqual(metadata.shape[1], 2)
        self.assertEqual(['Country', 'Subject', 'Frequency'], metadata.columns.names)
        indx = data.first_valid_index()
        sname = ('Albania', 'Population (Persons)', 'A')
        value = data.at[indx, sname]
        self.assertEqual(value, 2.762)

        indx = metadata.first_valid_index()
        value = metadata.at[indx, sname]
        self.assertEqual(value, '914')

        indx = data.last_valid_index()
        value = data.at[indx, sname]
        self.assertEqual(value, 2.858)

    def test_get_data_with_partial_selection_with_metadata_transform(self):
        """The method is testing getting series with partial selection"""

        _, metadata = knoema.get('IMFWEO2017Apr', True, subject='flibor6', frequency='A')
        self.assertEqual(metadata.shape[1], 2)
        self.assertEqual(['Country', 'Subject', 'Frequency'], metadata.columns.names)
        sname = ('Japan', 'Six-month London interbank offered rate (LIBOR) (Percent)', 'A')
        self.assertEqual(metadata.at['Country Id', sname], '158')
        self.assertEqual(metadata.at['Subject Id', sname], 'FLIBOR6')
        self.assertEqual(metadata.at['Unit', sname], 'Percent')


    def test_get_data_with_partial_selection_with_metadata(self):
        """The method is testing getting series with partial selection"""

        _, metadata = knoema.get('IMFWEO2017Apr', True, subject = 'flibor6')
        self.assertEqual(metadata.shape[1], 2)
        self.assertEqual(['Country', 'Subject', 'Frequency'], metadata.columns.names)
        sname = ('Japan', 'Six-month London interbank offered rate (LIBOR) (Percent)', 'A')
        self.assertEqual(metadata.at['Country Id',sname],'158')
        self.assertEqual(metadata.at['Subject Id',sname],'FLIBOR6')
        self.assertEqual(metadata.at['Unit',sname],'Percent')


    def test_get_data_with_partial_selection(self):
        """The method is testing getting series with partial selection"""

        data_frame = knoema.get('IMFWEO2017Apr', subject = 'flibor6')
        self.assertEqual(data_frame.shape[1], 2)
        self.assertEqual(['Country', 'Subject', 'Frequency'], data_frame.columns.names)

        indx = data_frame.first_valid_index()
        sname = ('Japan', 'Six-month London interbank offered rate (LIBOR) (Percent)', 'A')
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 10.861)

        indx = data_frame.index[38]
        value = data_frame.at[indx, sname]
        self.assertEqual(value, 0.048)

    def test_get_data_from_flat_dataset_without_time_and_with_metadata(self):
        """The method is testing load data from flat dataset without time and with metadata"""

        data_frame, metadata = knoema.get('pocqwkd', True, **{'Object type': 'Airports',
                                                              'Object name': 'Bakel airport'})

        self.assertEqual(data_frame.shape[0], 1)
        self.assertEqual(data_frame.shape[1], 6)
        self.assertEqual(metadata, None)

        self.assertEqual(data_frame.at[0, 'Latitude'], '14.847256')
        self.assertEqual(data_frame.at[0, 'Longitude'], '-12.468264')


    def test_get_data_from_flat_dataset_without_time(self):
        """The method is testing load data from flat dataset without time"""

        data_frame = knoema.get('pocqwkd', **{'Object type': 'Airports',
                                              'Object name': 'Bakel airport'})

        self.assertEqual(data_frame.shape[0], 1)
        self.assertEqual(data_frame.shape[1], 6)

        value = data_frame.at[0, 'Place']
        self.assertEqual(value, 'Bakel')

    def test_get_data_from_flat_dataset_with_multi_measures_and_metadata(self):
        """The method is testing load data from flat dataset with with mulitple measures and metadata"""

        data_frame, metadata = knoema.get('bmlaaaf', True, **{'Country': 'Albania',
                                              'Borrower': 'Ministry of Finance',
                                              'Guarantor': 'Albania',
                                              'Loan type': 'B loan',
                                              'Loan status': 'EFFECTIVE',
                                              'Currency of commitment': 'eur',
                                              'measure': 'Interest rate'})

        self.assertEqual(data_frame.shape[0], 1)
        self.assertEqual(data_frame.shape[1], 33)
        self.assertEqual(metadata, None)

        value = data_frame.at[0, 'Undisbursed Amount']
        self.assertEqual(value, 79998000.0)

    def test_get_data_from_flat_dataset_with_multi_measures(self):
        """The method is testing load data from flat dataset with with mulitple measures"""

        data_frame = knoema.get('bmlaaaf', **{'Country': 'Albania',
                                              'Borrower': 'Ministry of Finance',
                                              'Guarantor': 'Albania',
                                              'Loan type': 'B loan',
                                              'Loan status': 'EFFECTIVE',
                                              'Currency of commitment': 'eur',
                                              'measure': 'Interest rate'})

        self.assertEqual(data_frame.shape[0], 1)
        self.assertEqual(data_frame.shape[1], 33)

        value = data_frame.at[0, 'Interest Rate']
        self.assertEqual(value, 0.0)

    def test_get_data_from_flat_dataset(self):
        """The method is testing load data from flat dataset"""

        data_frame = knoema.get('cblymmf', Country='Albania;Australia', Keyword='FGP;TWP;TRP')
        self.assertEqual(data_frame.shape[0], 32)
        self.assertEqual(data_frame.shape[1], 4)

        self.assertAlmostEqual(float(data_frame.at[30, 'Value']), 98.8368, 4)

    def test_get_data_from_dataset_with_multiword_dimnames_and_metadata_transform(self):
        """The method is testing load data from regular dataset with dimenions that have multi word names include metadata"""

        data_frame, _ = knoema.get('FDI_FLOW_CTRY', True, **{'Reporting country': 'AUS',
                                                             'Partner country/territory': 'w0',
                                                             'Measurement principle': 'DI',
                                                             'Type of FDI': 'T_FA_F',
                                                             'Type of entity': 'ALL',
                                                             'Accounting entry': 'NET',
                                                             'Level of counterpart': 'IMC',
                                                             'Currency': 'USD',
                                                             'Frequency': 'A'})

        self.assertEqual(data_frame.shape[0], 7)
        self.assertEqual(data_frame.shape[1], 1)

        sname = (
        'Australia', 'WORLD', 'Directional principle: Inward', 'FDI financial flows - Total', 'All resident units',
        'Net', 'Immediate counterpart (Immediate investor or immediate host)', 'US Dollar', 'A')

        indx = data_frame.first_valid_index()
        value = data_frame.at[indx, sname]
        self.assertAlmostEqual(value, 31666.667, 3)

        indx = data_frame.last_valid_index()
        value = data_frame.at[indx, sname]
        self.assertAlmostEqual(value, 22267.638, 3)

    def test_get_data_from_dataset_with_multiword_dimnames_and_metadata(self):
        """The method is testing load data from regular dataset with dimenions that have multi word names include metadata"""

        data_frame, _ = knoema.get('FDI_FLOW_CTRY', True, **{'Reporting country': 'AUS',
                                                             'Partner country/territory': 'w0',
                                                             'Measurement principle': 'DI',
                                                             'Type of FDI': 'T_FA_F',
                                                             'Type of entity': 'ALL',
                                                             'Accounting entry': 'NET',
                                                             'Level of counterpart': 'IMC',
                                                             'Currency': 'USD'})

        self.assertEqual(data_frame.shape[0], 7)
        self.assertEqual(data_frame.shape[1], 1)

        sname = (
        'Australia', 'WORLD', 'Directional principle: Inward', 'FDI financial flows - Total', 'All resident units',
        'Net', 'Immediate counterpart (Immediate investor or immediate host)', 'US Dollar', 'A')

        indx = data_frame.first_valid_index()
        value = data_frame.at[indx, sname]
        self.assertAlmostEqual(value, 31666.667, 3)

        indx = data_frame.last_valid_index()
        value = data_frame.at[indx, sname]
        self.assertAlmostEqual(value, 22267.638, 3)

    def test_get_data_from_dataset_with_multiword_dimnames(self):
        """The method is testing load data from regular dataset with dimenions that have multi word names"""

        data_frame = knoema.get('FDI_FLOW_CTRY', **{'Reporting country': 'AUS',
                                                    'Partner country/territory': 'w0',
                                                    'Measurement principle': 'DI',
                                                    'Type of FDI': 'T_FA_F',
                                                    'Type of entity': 'ALL',
                                                    'Accounting entry': 'NET',
                                                    'Level of counterpart': 'IMC',
                                                    'Currency': 'USD'})

        self.assertEqual(data_frame.shape[0], 7)
        self.assertEqual(data_frame.shape[1], 1)

        sname = ('Australia', 'WORLD', 'Directional principle: Inward', 'FDI financial flows - Total', 'All resident units', 'Net', 'Immediate counterpart (Immediate investor or immediate host)', 'US Dollar', 'A')

        indx = data_frame.first_valid_index()
        value = data_frame.at[indx, sname]
        self.assertAlmostEqual(value, 31666.667, 3)

        indx = data_frame.last_valid_index()
        value = data_frame.at[indx, sname]
        self.assertAlmostEqual(value, 22267.638, 3)

    def test_get_data_from_dataset_by_dim_ids_transform(self):
        """The method is testing load data from regular dataset with dimenions that have multi word names by dim ids"""

        data_frame = knoema.get('FDI_FLOW_CTRY', **{'Reporting-country': 'AUS',
                                                    'Partner-country': 'w0',
                                                    'Measurement-principle': 'DI',
                                                    'Type-of-FDI': 'T_FA_F',
                                                    'Type-of-entity': 'ALL',
                                                    'Accounting-entry': 'NET',
                                                    'Level-of-counterpart': 'IMC',
                                                    'Currency': 'USD',
                                                    'Frequency': 'A'})

        self.assertEqual(data_frame.shape[0], 7)
        self.assertEqual(data_frame.shape[1], 1)

        sname = ('Australia', 'WORLD', 'Directional principle: Inward', 'FDI financial flows - Total', 'All resident units', 'Net', 'Immediate counterpart (Immediate investor or immediate host)', 'US Dollar', 'A')

        indx = data_frame.first_valid_index()
        value = data_frame.at[indx, sname]
        self.assertAlmostEqual(value, 31666.667, 3)

        indx = data_frame.last_valid_index()
        value = data_frame.at[indx, sname]
        self.assertAlmostEqual(value, 22267.638, 3)

    def test_get_data_from_dataset_by_dim_ids(self):
        """The method is testing load data from regular dataset with dimenions that have multi word names by dim ids"""

        data_frame = knoema.get('FDI_FLOW_CTRY', **{'Reporting-country': 'AUS',
                                                    'Partner-country': 'w0',
                                                    'Measurement-principle': 'DI',
                                                    'Type-of-FDI': 'T_FA_F',
                                                    'Type-of-entity': 'ALL',
                                                    'Accounting-entry': 'NET',
                                                    'Level-of-counterpart': 'IMC',
                                                    'Currency': 'USD'})

        self.assertEqual(data_frame.shape[0], 7)
        self.assertEqual(data_frame.shape[1], 1)

        sname = (
        'Australia', 'WORLD', 'Directional principle: Inward', 'FDI financial flows - Total', 'All resident units',
        'Net', 'Immediate counterpart (Immediate investor or immediate host)', 'US Dollar', 'A')

        indx = data_frame.first_valid_index()
        value = data_frame.at[indx, sname]
        self.assertAlmostEqual(value, 31666.667, 3)

        indx = data_frame.last_valid_index()
        value = data_frame.at[indx, sname]
        self.assertAlmostEqual(value, 22267.638, 3)


"""This module contains data definitions for Knoema client"""

from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta

import pandas
import knoema.api_definitions as definition
import knoema.view_definitions as view_definition

class DataReader(object):
    """This class read data from Knoema and transform it to pandas frame"""

    def __init__(self, client):
        self.client = client
        self.dataset = None
        self.include_metadata = False
        self.columns = None
        self.dimensions = []     
        self.separator = ';'

    def _get_series_name(self, series_point):
        names = []
        for dim in self.dimensions:
            names.append(series_point[dim.id])
        names.append(series_point['Frequency'])
        return tuple(names) 

    def _get_series_with_metadata(self, series_point):
        names = []
        for dim in self.dimensions:
            for item in dim.items:
                if item.name == series_point[dim.id]:
                    dim_attrs = item.fields
                    break
            for attr in dim.fields:          
                if not attr['isSystemField']: 
                    value_empty = True
                    for key, value in dim_attrs.items(): 
                        if definition.is_equal_strings_ignore_case(key, attr['name']):
                            names.append(value)
                            value_empty = False
                    if value_empty:
                        names.append(float("NaN"))
        names.append(series_point.get('Unit'))
        names.append(series_point.get('Scale'))
        names.append(series_point.get('Mnemonics'))
        for attr in self.dataset.timeseries_attributes:
            names.append(series_point.get(attr.name))  
        return tuple(names)     

    def _get_attribute_names(self):
        names = []
        for dim in self.dimensions:
            for attr in dim.fields:
                if not attr['isSystemField']:
                    names.append(dim.name +' '+ attr['displayName'])
        names.append('Unit')            
        names.append('Scale') 
        names.append('Mnemonics')
        for attr in self.dataset.timeseries_attributes:
            names.append(attr.name)     
        return names

    def _get_dimension_names(self):
        names = []
        for dim in self.dataset.dimensions:
            names.append(dim.name)
        if self.dataset.has_time:
            names.append('Frequency')             
        return names   

    def _get_metadata_series(self, resp, names_of_attributes):
        series = {}
        for series_point in resp.tuples:
            val = series_point['Value']
            if val is None:
                continue
            serie_name = self._get_series_name(series_point)
            if serie_name not in series:
                serie_attrs = self._get_series_with_metadata(series_point)
                series[serie_name] = KnoemaSeries(serie_name, serie_attrs, names_of_attributes, None)
        return series

    def _load_dimensions(self):
        for dim in self.dataset.dimensions:
            self.dimensions.append(self.client.get_dimension(self.dataset.id, dim.id))
    

class SelectionDataReader(DataReader):

    def __init__(self, client, dim_values, transform = None):
        super().__init__(client)
        self.dim_values = dim_values
        self.transform = transform

    def _get_dim_members(self, dim, splited_values):
        members = []
        for value in splited_values:
            if value is None or isinstance(value, str) and not value:
                raise ValueError('Selection for dimension {} is empty'.format(dim.name))

            member = dim.find_member_by_id(value)
            if member is None:
                member = dim.find_member_by_name(value)

            if member is None:
                member = dim.find_member_by_regionid(value)

            if member is None:
                member = dim.find_member_by_ticker(value)

            if member is None and value.isnumeric():
                member = dim.find_member_by_key(int(value))

            if member:
                members.append(member.key)
            else:
                raise ValueError('Selection for dimension {} contains invalid elements'.format(dim.name))

        return members    

    def _find_dimension(self, dim_name_or_id):

        dim = self.dataset.find_dimension_by_name(dim_name_or_id)
        if dim is None:
            dim = self.dataset.find_dimension_by_id(dim_name_or_id)

        return dim

    def _add_full_selection_by_empty_dim_values(self, filter_dims, pivot_req):
        out_of_filter_dim_id = [dim.id for dim in self.dataset.dimensions if dim not in filter_dims]
        for id in out_of_filter_dim_id:
            pivot_req.stub.append(definition.PivotItem(id, []))

    def _create_pivot_request(self):

        pivot_req = definition.PivotRequest(self.dataset.id)

        filter_dims = []
        time_range = None
        for name, value in self.dim_values.items():
            if definition.is_equal_strings_ignore_case(name, 'timerange'):
                time_range = value
                continue

            splited_values = value.split(self.separator) if isinstance(value, str) else value
            if definition.is_equal_strings_ignore_case(name, 'frequency'):
                pivot_req.frequencies = splited_values
                continue

            dim = self._find_dimension(name)
            if dim is None:
                raise ValueError('Dimension with id or name {} is not found'.
                                 format(name))

            filter_dims.append(dim)

            for dimension in self.dimensions:
                if dimension.id == dim.id:
                    dim = dimension
                    break

            aggregations = None
            if len(splited_values) > 0 and splited_values[0].startswith('@'):
                aggregations = splited_values[0][1:]
                splited_values = splited_values[1:]

            members = self._get_dim_members(dim, splited_values)
            if not members and aggregations == None:
                raise ValueError('Selection for dimension {} is empty'.format(dim.name))

            pivot_req.stub.append(definition.PivotItem(dim.id, members, aggregations = aggregations))

        self._add_full_selection_by_empty_dim_values(filter_dims, pivot_req)

        if time_range:
            pivot_req.header.append(definition.PivotTimeItem('Time', [time_range], 'range'))
        else:
            pivot_req.header.append(definition.PivotTimeItem('Time', [], 'AllData'))

        if self.transform is not None:
            pivot_req.transform = self.transform

        if self.columns is not None:
            pivot_req.columns = self.columns

        return pivot_req

class ResponseReader(object):
    def __init__(self, reader):
        self.include_metadata = reader.include_metadata
        self.dataset = reader.dataset
        self.reader = reader
        super().__init__()

    def _get_detail_columns(self, resp):
        detail_columns = []
        if resp.descriptor is not None and 'detailColumns' in resp.descriptor and resp.descriptor['detailColumns'] is not None:
            for column in resp.descriptor['detailColumns']:
                detail_columns.append(column['name'])

        return detail_columns if len(detail_columns) > 0 else None

class PivotResponseReader(ResponseReader):
    def __init__(self, reader, pivot_resp):
        self.pivot_resp = pivot_resp
        super().__init__(reader)
    
    def get_pandasframe(self):
        names_of_dimensions = self.reader._get_dimension_names()
            
        # create dataframe with data
        detail_columns = self._get_detail_columns(self.pivot_resp)
        series = self._get_data_series(self.pivot_resp, detail_columns)
        pandas_series = PandasHelper.creates_pandas_series(series, {}, detail_columns)
        pandas_data_frame = PandasHelper.create_pandas_dataframe(pandas_series, names_of_dimensions, detail_columns)
        if not self.include_metadata:
            return pandas_data_frame
            
        # create dataframe with metadata
        series_with_attr = self._get_metadata_series(self.pivot_resp)
        pandas_series_with_attr = PandasHelper.creates_pandas_series(series_with_attr, {}, None)
        pandas_data_frame_with_attr = PandasHelper.create_pandas_dataframe(pandas_series_with_attr, names_of_dimensions, None)
        return pandas_data_frame, pandas_data_frame_with_attr

    def _get_data_series(self, resp, detail_columns):
        series_map = {}

        columns = None
        frequency_list = []
        for series_point in resp.tuples:
            val = series_point['Value']
            if val is None:
                continue
            series_name = self._get_series_name(series_point)
            if series_name in series_map:
                series = series_map[series_name]
            else:
                series = KnoemaSeries(series_name, [], [], detail_columns)
                series_map[series_name] = series

            freq = series_point['Frequency']

            if 'Time' in series_point:
                curr_date_val = series_point['Time']
                try:
                    curr_date_val = datetime.strptime(series_point['Time'], '%Y-%m-%dT%H:%M:%SZ')

                    if (freq == "W"):
                        curr_date_val = curr_date_val - timedelta(days = curr_date_val.weekday())
                except ValueError:
                    pass
            else:
                curr_date_val = 'All time'

            if freq not in frequency_list:
                frequency_list.append(freq)
            if freq == "W":
                curr_date_val = curr_date_val - timedelta(days = curr_date_val.weekday())
            if freq == "FQ":
                curr_date_val = TimeFormat.format_statistical(curr_date_val, 'FQ')
            if detail_columns is not None:
                columns = []
                for column_name in detail_columns:
                    columns.append(series_point[column_name])

            series.add_value(series_point['Value'], curr_date_val, columns)

        if 'FQ' in frequency_list and len(frequency_list) > 1:
            raise ValueError('Please provide a valid frequency list. You can request FQ or others frequencies not together.')

        return series_map

    def _get_series_name(self, series_point):
        names = []
        for dim in self.dataset.dimensions:
            if dim.id in series_point:
                names.append(series_point[dim.id])
        if self.dataset.has_time:
            names.append(series_point['Frequency'])
        return tuple(names) 

    def _get_metadata_series(self, resp):
        series = {}
        names_of_attributes = self._get_attribute_names(resp)
        for series_point in resp.tuples:
            serie_name = self._get_series_name(series_point)
            if serie_name not in series:
                serie_attrs = self._get_series_with_metadata(series_point, resp)
                series[serie_name] = KnoemaSeries(serie_name, serie_attrs, names_of_attributes, None)
        return series  

    def _get_series_with_metadata(self, series_point, resp):
        names = []
        resp_dims = [dim for dim in (resp.header + resp.stub + resp.filter) if dim.fields and any(dim.fields)]
        for dim in resp_dims:
            if dim.dimensionid == 'Time' or not dim.metadataFields:
                continue
            for item in dim.metadataFields:
                if item.name == series_point[dim.dimensionid]:
                    dim_attrs = item.fields
                    break

            for attr in dim.fields: 
                find_elem = False
                if attr['isSystemField']:
                    continue
                for key, value in dim_attrs.items(): 
                    if definition.is_equal_strings_ignore_case(key, attr['name']):
                        find_elem = True
                        names.append(value)
                        break
                if not find_elem:
                    names.append(None)

        names.append(series_point.get('Unit'))
        names.append(series_point.get('Scale'))
        names.append(series_point.get('Mnemonics'))
        for attr in self.dataset.timeseries_attributes:
            names.append(series_point.get(attr.name))  
        return tuple(names)     
        
    def _get_attribute_names(self, resp):
        resp_dims = [dim for dim in (resp.header + resp.stub + resp.filter) if dim.fields and any(dim.fields)]
        names = []
        for dim in resp_dims:
            for d in self.dataset.dimensions:
                if dim.dimensionid == d.id:
                    for attr in dim.fields:
                        if not attr['isSystemField']:
                            names.append(d.name +  ' ' + attr["displayName"])
                    break
        names.append('Unit')            
        names.append('Scale') 
        names.append('Mnemonics')
        for attr in self.dataset.timeseries_attributes:
            names.append(attr.name)     
        return names

class StreamingResponseReader(ResponseReader):
    def __init__(self, reader, data_streaming):
        self.data_streaming = data_streaming
        super().__init__(reader)
    
    def get_pandasframe(self):
        detail_columns = self._get_detail_columns(self.data_streaming)
        # create dataframe with data
        series = self._get_data_series(self.data_streaming, detail_columns)

        pandas_series = {}
        names_of_dimensions = self.reader._get_dimension_names()
        if self.include_metadata:
            pandas_series_with_attr = {}
            names_of_attributes = self._get_attribute_names()

        pandas_series = PandasHelper.creates_pandas_series(series, pandas_series, detail_columns)
        pandas_data_frame = PandasHelper.create_pandas_dataframe(pandas_series, names_of_dimensions, detail_columns)
        if not self.include_metadata:
            return pandas_data_frame
            
        # create dataframe with metadata
        series_with_attr = self._get_metadata_series(self.data_streaming, names_of_attributes)
        pandas_series_with_attr = PandasHelper.creates_pandas_series(series_with_attr, pandas_series_with_attr, None)
        pandas_data_frame_with_attr = PandasHelper.create_pandas_dataframe(pandas_series_with_attr, names_of_dimensions, None)         
        return pandas_data_frame, pandas_data_frame_with_attr

    def _get_data_series(self, resp, detail_columns):
        series_map = {}
        dict_with_delta = TimeFormat.get_frequencies_delta()

        frequency_list = []
        use_stat_format_for_date_label = self._get_use_stat_format_for_date_label(resp.series)
        dimensions_members_with_id = self._get_dimensions_members_with_id(resp.series)

        detail_values = None
        for series_point in resp.series:  
            all_values = series_point['values']  
            series_name = self._get_series_name(series_point, dimensions_members_with_id)
            if detail_columns is not None:
                detail_values = []
                for column_name in detail_columns:
                    detail_values.append(series_point[column_name])

            date_format = '%Y-%m-%dT%H:%M:%S' + ('Z' if series_point['startDate'].endswith('Z') else '')
            data_begin_val = datetime.strptime(series_point['startDate'], date_format)
            freq = series_point['frequency']
            if freq not in frequency_list:
                frequency_list.append(freq)
            if (freq == "W"):
                data_begin_val = data_begin_val - timedelta(days = data_begin_val.weekday())
            
            delta = dict_with_delta[freq]
            series = KnoemaSeries(series_name, [], [], detail_columns)
            date_labels = series_point['dateLabels'] if 'dateLabels' in series_point else None
            curr_date_val = data_begin_val
            for vi in range(0, len(all_values)):
                val = all_values[vi]
                if val is not None:
                    date = curr_date_val if freq != 'FQ' else TimeFormat.format_statistical(curr_date_val, 'FQ')
                    if freq in use_stat_format_for_date_label or date_labels is not None and date_labels[vi] is not None:
                        date = TimeFormat.format_statistical(curr_date_val, freq) \
                            if freq in use_stat_format_for_date_label else datetime.strptime(date_labels[vi], date_format)
                    series.index.append(date)
                    series.values.append(val)
                    for ai in range(0, series.column_count):
                        series.column_values[ai].append(detail_values[ai][vi])
                curr_date_val += delta

            series_map[series_name] = series

        if 'FQ' in frequency_list and len(frequency_list) > 1:
            raise ValueError('Please provide a valid frequency list. You can request FQ or others frequencies not together.')

        return series_map

    def _get_use_stat_format_for_date_label(self, series):
        use_stat_format_for_date_label = {}
        date_labels_by_freq = {}
        has_date_labels_by_freq = {}
        dict_with_delta = TimeFormat.get_frequencies_delta()
        for series_point in series:  
            freq = series_point['frequency']
            if freq in use_stat_format_for_date_label:
                continue
            has_date_labels = 'dateLabels' in series_point
            if freq in has_date_labels_by_freq and has_date_labels_by_freq[freq] != has_date_labels:
                use_stat_format_for_date_label[freq] = True
                continue
            has_date_labels_by_freq[freq] = has_date_labels
            if not has_date_labels:
                continue

            if freq not in date_labels_by_freq:
                date_labels_by_freq[freq] = {}

            date_labels = series_point['dateLabels']

            all_values = series_point['values']  

            date_format = '%Y-%m-%dT%H:%M:%S' + ('Z' if series_point['startDate'].endswith('Z') else '')
            data_begin_val = datetime.strptime(series_point['startDate'], date_format)
            if (freq == "W"):
                data_begin_val = data_begin_val - timedelta(days = data_begin_val.weekday())
            
            delta = dict_with_delta[freq]
            curr_date_val = data_begin_val
            for vi in range(0, len(all_values)):
                if curr_date_val in date_labels_by_freq[freq]:
                    if date_labels[vi] is not None and date_labels_by_freq[freq][curr_date_val] != date_labels[vi]:
                        use_stat_format_for_date_label[freq] = True
                        break
                else:
                    date_labels_by_freq[freq][curr_date_val] = date_labels[vi]
                curr_date_val += delta
        return use_stat_format_for_date_label

    def _get_dimensions_members_with_id(self, series):
        dimensions_members_with_id = {}

        for dim in self.dataset.dimensions:
            names = set()
            names_with_id = set()
            ids = set()
            for series_point in series:
                name = series_point[dim.id]['name'] if 'name' in series_point[dim.id] else series_point[dim.id]
                id = series_point[dim.id]['id'] if 'id' in series_point[dim.id] else str(series_point[dim.id])

                if (name in names and id not in ids):
                    names_with_id.add(name)
                else:
                    names.add(name)

                ids.add(id)

            if (len(names_with_id) > 0):
                dimensions_members_with_id[dim.id] = names_with_id


        return dimensions_members_with_id

    def _get_series_name(self, series_point, dimensions_members_with_id):
        names = []
        for dim in self.dataset.dimensions:
            name = series_point[dim.id]['name'] if 'name' in series_point[dim.id] else series_point[dim.id]
            members_with_id = dimensions_members_with_id[dim.id] if dim.id in dimensions_members_with_id else []
            if (name in members_with_id):
                id = series_point[dim.id]['id'] if 'id' in series_point[dim.id] else series_point[dim.id]
                name = name + '(' + id + ')'
            names.append(name)
        if 'frequency' in series_point:
            names.append(series_point['frequency'])
        return tuple(names)

    def _get_attribute_names(self):
        names = []
        for dim in self.dataset.dimensions:
            for attr in self.data_streaming.dimensionFields[dim.id]:
                if not attr['isSystemField']:
                    names.append(dim.name +' '+ attr['displayName'])
        names.append('Unit')            
        names.append('Scale') 
        names.append('Mnemonics')
        for attr in self.dataset.timeseries_attributes:
            names.append(attr.name)     
        return names

    def _get_metadata_series(self, resp, names_of_attributes):
        series = {}
        dimensions_members_with_id = self._get_dimensions_members_with_id(resp.series)
        for series_point in resp.series:
            serie_name = self._get_series_name(series_point, dimensions_members_with_id)
            if serie_name not in series:
                serie_attrs = self._get_series_with_metadata(series_point, resp.dimensionFields)
                series[serie_name] = KnoemaSeries(serie_name, serie_attrs, names_of_attributes, None)
        return series  

    def _get_series_with_metadata(self, series_point, dimensionFields):
        names = []
        for dim in self.dataset.dimensions:
            dimFields = dimensionFields[dim.id]
            dimValues = series_point[dim.id]
            for attr in dimFields: 
                if not attr['isSystemField']: 
                    if attr['name'] in dimValues:
                        names.append(dimValues[attr['name']])
                    else:
                        names.append(None)        

        names.append(series_point.get('unit'))
        names.append(series_point.get('scale'))
        names.append(series_point.get('mnemonics'))
        for attr in self.dataset.timeseries_attributes:
            names.append(series_point['timeseriesAttributes'][attr.name])
        return tuple(names)

class DetailsResponseReader(ResponseReader):
    def __init__(self, reader, data_resp):
        self.data_resp = data_resp
        super().__init__(reader)
    
    def get_pandasframe(self):
        records = self.convert_pandasframe()
        if not self.include_metadata:
            return records
        return records, None

    def convert_pandasframe(self):
        titles = []
        columns = []
        indexes = []
        for col in self.data_resp.columns:
            indexes.append(col['index'])
            if col['dimensionId'] != None:
                if 'detailColumns' in col:
                    titles.append(col['name'])
                    columns.append([col['id'], 'name'])
                    for detail in col['detailColumns']:
                        indexes.append(detail['index'])
                        titles.append(detail['name'])
                        columns.append([col['id'], detail['id']])
                else:
                    titles.append(col['name'])
                    columns.append(col['id'])
            else:
                if col['type'] == 'Date':
                    titles.append(col['name'])
                    columns.append([col['id'], 'value'])
                    continue

                if col['type'] == 'Currency':
                    titles.append(col['name'])
                    columns.append([col['id'], 'value'])
                    continue

                titles.append(col['name'])
                columns.append(col['id'])

        #sort both array by indexes
        titles, columns = zip(*[x for _,x in sorted(zip(indexes, zip(titles, columns)))])
        records = []
        for tuple in self.data_resp.tuples:
            record = []

            for col in columns:
                if isinstance(col, str):
                    val = tuple[col]
                    record.append(val)
                else:
                    val = tuple[col[0]]
                    record.append(val[col[1]] if val != None else None)

            records.append(record)

        return pandas.DataFrame(data=records, columns=titles)        

class TransformationDataReader(SelectionDataReader):

    def __init__(self, client, dim_values, transform, frequency):
        if dim_values == None:
            dim_values = {}
        if transform:
            dim_values['transform'] = transform
        if frequency:
            dim_values['frequency'] = frequency
        super().__init__(client, dim_values)
        
    def get_pandasframe(self):
        data_resp = self.client.get_dataset_data(self.dataset.id, self._get_data_filters())
        if isinstance(data_resp, definition.DetailsResponse):
            response_reader = DetailsResponseReader(self, data_resp)
            return response_reader.get_pandasframe()

        if isinstance(data_resp, definition.RawDataResponse):
            token = data_resp.continuation_token
            while token is not None:
                res2 = self.client.get_data_raw_with_token(token)
                data_resp.series += res2.series
                token = res2.continuation_token

            response_reader = StreamingResponseReader(self, data_resp)
            return response_reader.get_pandasframe()

        response_reader = PivotResponseReader(self, data_resp)
        return response_reader.get_pandasframe()

    def _get_data_filters(self):
        filter_dims = {}
        passed_params = ['timerange', 'transform', 'timesince', 'timelast', 'timemembers']

        for name, value in self.dim_values.items():
            if name.lower() in passed_params:
                filter_dims[name] = value
                continue
            if definition.is_equal_strings_ignore_case(name, 'datecolumn') and self.dataset.type != 'Regular':
                filter_dims['datecolumn'] = value
                continue

            splited_values = [x for x in value.split(self.separator) if x] if isinstance(value, str) else value
            if definition.is_equal_strings_ignore_case(name, 'frequency'):
                filter_dims["frequency"] = self.separator.join(splited_values)
                continue

            dim = self._find_dimension(name)
            if dim is None:
                raise ValueError('Dimension with id or name {} is not found'.
                                 format(name))

            if not splited_values:
                raise ValueError('Selection for dimension {} is empty'.format(dim.name))

            filter_dims[dim.id] =  self.separator.join(s for s in splited_values)

        if self.include_metadata:
            filter_dims['metadata'] = 'true'
        
        if self.separator != ',':
             filter_dims['separator'] = self.separator

        return definition.DataAPIRequest(filter_dims)

class StreamingDataReader(SelectionDataReader):

    def __init__(self, client, dim_values, transform = None, group_by = None):
        self.group_by = group_by

        super().__init__(client, dim_values, transform)
    
    def get_series_metadata(self):
        self._load_dimensions()
        pivot_req = self._create_pivot_request()

        data_streaming = self.client.get_data_raw(pivot_req, True)
        
        return data_streaming.series

    def get_pandasframe(self):
        self._load_dimensions()
        pivot_req = self._create_pivot_request()
        data_streaming = self.client.get_data_raw(pivot_req)
        response_reader = StreamingResponseReader(self, data_streaming)
        return response_reader.get_pandasframe()

    def _get_series_with_attr(self, series, series_with_attr):
        res = {}
        for series_name, _ in series.items():
            if series_name in series_with_attr:
                res[series_name] = series_with_attr[series_name]

        return res

    def _get_frequency_for_normalization(self, required_frequency, available_frequency):
        sorted_frequency = ['A', 'H', 'Q', 'M', 'W', 'D']
        sorted_available_frequency = []
        for f in sorted_frequency:
            if f in available_frequency:
                sorted_available_frequency.append(f)

        norm_index = sorted_frequency.index(sorted_available_frequency[-1])
        for f in sorted_available_frequency:
            if sorted_frequency.index(f) > sorted_frequency.index(required_frequency):
                norm_index = sorted_frequency.index(f)
                break

        return sorted_frequency[norm_index]

    def _get_all_series_for_group(self, group_name, group_name_index, frequency, series, metadata):
        series_by_group = {}
        for series_name, series_item in series.items():
            if series_name[group_name_index] == group_name:
                series_by_group[series_name] = series_item
        
        available_frequency = {}
        count = 0
        for md in metadata:
            if md[self.group_by]['name'] == group_name:
                freq = md['frequency']
                if freq not in available_frequency:
                    available_frequency[freq] = 0
                    
                available_frequency[freq] += 1
                count += 1

        if frequency == None:
            if len(series_by_group.keys()) == count:
                return series_by_group
            else:
                return None

        if len(available_frequency.keys()) == 0:
            return None

        norm_frequency = frequency
        if frequency not in available_frequency.keys():
            norm_frequency = self._get_frequency_for_normalization(frequency, available_frequency)
        
        if len(series_by_group.keys()) == available_frequency[norm_frequency]:
                return series_by_group

        return None

    def get_pandasframe_by_metadata_grouped(self, metadata, frequency, timerange):
        self._load_dimensions()
        names_of_dimensions = self._get_dimension_names()
        
        series = {}
        series_with_attr = {}
        offset = 0
        detail_columns = None
        was = False
        while True:
            metadata_part = self._get_part_of_metadata(metadata, offset)
            if metadata_part == None:
                break

            offset += len(metadata_part)

            dim_values = {}
            for item in metadata_part:
                for dim in self.dataset.dimensions:
                    dim_id = dim.id

                    if dim_id not in item:
                        raise ValueError('There is no value for dim: {}'.format(dim_id))

                    member_key = str(item[dim_id]['key'])
                    if dim_id in dim_values:
                        if member_key not in dim_values[dim_id]:
                            dim_values[dim_id].append(member_key)
                    else:
                        dim_values[dim_id] = [member_key]

            if frequency != None:
                dim_values['frequency'] = frequency
            if timerange != None:
                dim_values['timerange'] = timerange

            self.dim_values = dim_values
            pivot_req = self._create_pivot_request()
            pivot_resp = self.client.get_data_raw(pivot_req)
            response_reader = StreamingResponseReader(self, pivot_resp)

            if was:
                if detail_columns is not None:
                    part_detail_columns = response_reader._get_detail_columns(pivot_resp)
                    if detail_columns != part_detail_columns:
                        detail_columns = None
            else:
                was = True
                detail_columns = response_reader._get_detail_columns(pivot_resp)

            part_series = response_reader._get_data_series(pivot_resp, detail_columns)
            series.update(part_series)

            if self.include_metadata:
                names_of_attributes = self._get_attribute_names()
                part_series_with_attr = response_reader._get_metadata_series(pivot_resp, names_of_attributes)
                series_with_attr.update(part_series_with_attr)

            group_name_index = -1
            for i in range(len(self.dataset.dimensions)):
                if self.group_by == self.dataset.dimensions[i].id or self.group_by == self.dataset.dimensions[i].name:
                    self.group_by = self.dataset.dimensions[i].id
                    group_name_index = i
                    break

            groups_to_delete = []
            groups_checked = []
            for series_name, series_item in series.items():
                group_name = series_name[group_name_index]

                if group_name in groups_to_delete:
                    continue
                if group_name in groups_checked:
                    continue

                all_series_by_group = self._get_all_series_for_group(group_name, group_name_index, frequency, series, metadata)
                if all_series_by_group != None:
                    groups_to_delete.append(group_name)
                    all_panda_series_by_group = PandasHelper.creates_pandas_series(all_series_by_group, {}, detail_columns)
                    data_frame = definition.DataFrame()
                    data_frame.id = group_name
                    data_frame.data = PandasHelper.create_pandas_dataframe(all_panda_series_by_group, names_of_dimensions, detail_columns)

                    if self.include_metadata:
                        all_series_with_attr_by_group = self._get_series_with_attr(all_series_by_group, series_with_attr)
                        all_pandes_series_with_attr_by_group = PandasHelper.creates_pandas_series(all_series_with_attr_by_group, {}, None)
                        data_frame.metadata = PandasHelper.create_pandas_dataframe(all_pandes_series_with_attr_by_group, names_of_dimensions, None)
                        
                    yield data_frame
                else:
                    groups_checked.append(group_name)

            left_series = {}
            for series_name, series_item in series.items():
                group_name = series_name[group_name_index]
                if group_name not in groups_to_delete:
                    left_series[series_name] = series_item
                
            series = left_series

            left_series_with_attr = {}
            for series_name, series_item in series_with_attr.items():
                group_name = series_name[group_name_index]
                if group_name not in groups_to_delete:
                    left_series_with_attr[series_name] = series_item
                
            series_with_attr = left_series_with_attr

    def _get_part_of_metadata(self, metadata, offset):
        limit = 50000
        member_count_limit = 200

        if offset >= len(metadata):
            return None

        curr_points = 0
        curr_member_count = 0
        time_points = 0
        dim_values = {}
        index = offset
        for i in range(offset, len(metadata)):
            curr_ts = metadata[i]
            curr_time_points = self._get_time_point_amount(curr_ts['startDate'], curr_ts['endDate'], curr_ts['frequency'])
            if curr_time_points > time_points:
                time_points = curr_time_points

            curr_points = time_points
            curr_member_count = 0
            for dim in self.dataset.dimensions:
                dim_id = dim.id

                if dim_id not in curr_ts:
                    raise ValueError('There is no value for dim: {}'.format(dim_id))

                member_key = str(curr_ts[dim_id]['key'])
                if dim_id in dim_values:
                    if member_key not in dim_values[dim_id]:
                        dim_values[dim_id].append(member_key)
                else:
                    dim_values[dim_id] = [member_key]

                curr_points = curr_points * len(dim_values[dim_id])
                curr_member_count = curr_member_count + len(dim_values[dim_id])

            if curr_points >= limit or curr_member_count >= member_count_limit:
                break

            index = i

        return metadata[offset:min(index + 1, len(metadata))]

    def _get_time_point_amount(self, start_date, end_date, frequency):
        count = 0
        dict_with_delta = TimeFormat.get_frequencies_delta()

        date_format = '%Y-%m-%dT%H:%M:%S' + ('Z' if start_date.endswith('Z') else '')
        data_begin_val = datetime.strptime(start_date, date_format)
        date_format = '%Y-%m-%dT%H:%M:%S' + ('Z' if end_date.endswith('Z') else '')
        data_end_val = datetime.strptime(end_date, date_format)
        if (frequency == "W"):
            data_begin_val = data_begin_val - timedelta(days = data_begin_val.weekday())
            data_end_val = data_end_val - timedelta(days = data_end_val.weekday())
        delta = dict_with_delta[frequency]
        curr_date_val = data_begin_val
        while curr_date_val<=data_end_val:
            curr_date_val += delta
            count += 1

        # we have to increase amount of points for FQ freq because of export from OASIS issue
        if frequency == 'FQ':
           count += 10
                
        return count

class MnemonicsDataReader(DataReader):

    def __init__(self, client, mnemonics, transform, frequency):
        super().__init__(client)
        self.mnemonics = mnemonics
        self.transform = transform
        self.frequency = frequency

    def _get_metadata_series(self, resp, names_of_attributes):
        series = {}
        for series_point in resp.tuples:
            val = series_point['Value']
            if val is None:
                continue
            serie_name = series_point['Mnemonics']
            if serie_name not in series:
                serie_attrs = self._get_series_with_metadata(series_point)
                series[serie_name] = KnoemaSeries(serie_name, serie_attrs, names_of_attributes, None)
        return series  

    def _get_detail_columns(self, resp):
        return None

    def _get_data_series(self, resp, detail_columns):
        series = {}
        frequency_list = []
        for series_point in resp.tuples:
            val = series_point['Value']
            if val is None:
                continue
            series_name = series_point['Mnemonics']
            if series_name not in series:
                series[series_name] = KnoemaSeries(series_name, [], [], detail_columns)

            curr_date_val = series_point['Time']
            try:
                curr_date_val = datetime.strptime(series_point['Time'], '%Y-%m-%dT%H:%M:%SZ')
            except ValueError:
                pass

            freq = series_point['Frequency']
            if freq not in frequency_list:
                frequency_list.append(freq)
            if (freq == "W"):
                curr_date_val = curr_date_val - timedelta(days = curr_date_val.weekday())
            if (freq == 'FQ'):
                curr_date_val = TimeFormat.format_statistical(curr_date_val, 'FQ')
            series[series_name].add_value(series_point['Value'], curr_date_val, None)

        if 'FQ' in frequency_list and len(frequency_list) > 1:
            raise ValueError('Please provide a valid frequency list. You can request FQ or others frequencies not together.')

        return series

    def _get_pandasframe_one_dataset(self):
        pandas_series = {}
        if self.include_metadata:
            pandas_series_with_attr = {}
            names_of_attributes = self._get_attribute_names()

        mnemonics = self.mnemonics
        mnemonics_string = self.separator.join(mnemonics) if isinstance(mnemonics, list) else mnemonics
        mnemonics_resp = self.client.get_mnemonics(mnemonics_string, self.transform, self.frequency)
        detail_columns = None
            
        for item in mnemonics_resp.items:
            pivot_resp = item.pivot
            if not definition.is_equal_strings_ignore_case(self.dataset.id, pivot_resp.dataset):
                continue
            # create dataframe with data for mnemonics
            series = self._get_data_series(pivot_resp, detail_columns)
            pandas_series = PandasHelper.creates_pandas_series(series, pandas_series, detail_columns)
            if self.include_metadata:
                # create dataframe with metadata for mnemonics
                series_with_attr = self._get_metadata_series(pivot_resp, names_of_attributes)
                pandas_series_with_attr = PandasHelper.creates_pandas_series(series_with_attr, pandas_series_with_attr, None)

        pandas_data_frame = PandasHelper.create_pandas_dataframe(pandas_series, [], detail_columns)
        if not self.include_metadata:
            return pandas_data_frame
        pandas_data_frame_with_attr = PandasHelper.create_pandas_dataframe(pandas_series_with_attr, [], None)
        return pandas_data_frame, pandas_data_frame_with_attr

    def _get_pandasframe_across_datasets(self):
           
        mnemonics_string = self.separator.join(self.mnemonics) if isinstance(self.mnemonics, list) else self.mnemonics
        mnemonics_resp = self.client.get_mnemonics(mnemonics_string, self.transform, self.frequency)

        dict_datasets = {}
        pandas_series = {}
        if self.include_metadata:
            pandas_series_with_attr = {}

        detail_columns = None
        for item in mnemonics_resp.items:
            pivot_resp = item.pivot
            if pivot_resp is None:
                continue
            dataset_id = pivot_resp.dataset
            if dataset_id  not in dict_datasets:
                dataset = self.client.get_dataset(dataset_id)
                self.dataset = dataset
                dimensions = []
                for dim in dataset.dimensions:
                    dimensions.append(self.client.get_dimension(dataset_id, dim.id))
                self.dimensions = dimensions
                names_of_attributes = self._get_attribute_names() if self.include_metadata else None
                dict_datasets[dataset_id] = (dataset, dimensions, names_of_attributes)
            else:
                self.dataset, self.dimensions, names_of_attributes = dict_datasets[dataset_id]
                    
            # create dataframe with data for mnemonics
            series = self._get_data_series(pivot_resp, detail_columns)
            pandas_series = PandasHelper.creates_pandas_series(series, pandas_series, detail_columns)
            if self.include_metadata:
                # create dataframe with metadata for mnemonics
                series_with_attr = self._get_metadata_series(pivot_resp, names_of_attributes)
                pandas_series_with_attr = PandasHelper.creates_pandas_series(series_with_attr, pandas_series_with_attr, None)

        pandas_data_frame = PandasHelper.create_pandas_dataframe(pandas_series, [], detail_columns)
        if not self.include_metadata:
            return pandas_data_frame
        pandas_data_frame_with_attr = PandasHelper.create_pandas_dataframe(pandas_series_with_attr, [], None)         
        return pandas_data_frame, pandas_data_frame_with_attr

    def get_pandasframe(self):
        """The method loads data from dataset"""
        if self.dataset:
            self._load_dimensions()
            return self._get_pandasframe_one_dataset()
        return self._get_pandasframe_across_datasets()

class KnoemaSeries(object):
    """This class combines values and index points for one time series"""

    def __init__(self, name, values=[], index=[], detail_columns=None):
        self.name = name
        self.values = values
        self.index = index
        self.column_count = len(detail_columns) if detail_columns is not None else 0
        self.column_values = [] if self.column_count > 0 else None
        for _ in range(0, self.column_count):
            self.column_values.append([])

    def add_value(self, value, index_point, columns=None):
        """The function is addeing new value to provied index. If index does not exist"""
        if index_point not in self.index:
            self.values.append(value)
            self.index.append(index_point)
            if self.column_count > 0:
                for i in range(0, self.column_count):
                    self.column_values[i].append(None if columns is None else columns[i])

    def creates_pandas_series(self, pandas_series, detail_columns):
        """The function creates pandas series based on index and values"""
        if detail_columns is None:
            pandas_series[self.name] = pandas.Series(self.values, self.index, name=self.name)
        else:
            series_name = self.name + ('Value',)
            pandas_series[series_name] = pandas.Series(self.values, self.index, name=series_name)
            for i in range(0, self.column_count):
                column_name = detail_columns[i]
                series_name = self.name + (column_name,)
                pandas_series[series_name] = pandas.Series(self.column_values[i], self.index, name=series_name)

class PandasHelper(object):
    @staticmethod
    def creates_pandas_series(series, pandas_series, detail_columns):
        for _, series_content in series.items():
            series_content.creates_pandas_series(pandas_series, detail_columns)
        return pandas_series

    @staticmethod
    def create_pandas_dataframe(pandas_series, names_of_dimensions, detail_columns):
        pandas_data_frame = pandas.DataFrame(pandas_series)
        pandas_data_frame.sort_index()
        if isinstance(pandas_data_frame.columns, pandas.MultiIndex):
            column_names = names_of_dimensions
            if detail_columns is not None:
                column_names = list(column_names)
                column_names.append('Attribute')
            pandas_data_frame.columns.names = column_names

        return pandas_data_frame

class TimeFormat(object):
    @staticmethod
    def format_statistical(date_point, freq):
        return {
            'FQ': lambda d: '{}FQ{}'.format(d.year, (d.month - 1) // 3 + 1),
            'W': lambda d: TimeFormat.format_weekly(d)
        }.get(freq, date_point)(date_point)

    @staticmethod
    def format_weekly(date):
        iso_values = date.isocalendar()
        iso_year = iso_values[0]
        week_number = iso_values[1]
        return '{}W{}'.format(iso_year, week_number)

    @staticmethod
    def get_frequencies_delta():
        return {
            'A': relativedelta(years = 1),
            'H': relativedelta(months = 6),
            'Q': relativedelta(months = 3),
            'FQ': relativedelta(months = 3),
            'M': relativedelta(months = 1),
            'W': timedelta(days = 7),
            'D': timedelta(days = 1)}


class DimensionMetadataReader:

    def __init__(self, client, dataset, dimension):
        self.client = client
        self.dataset = dataset
        self.dimension = dimension

    def get(self):
        dim_info = self.client.get_dimension(self.dataset, self.dimension)
        view_model = view_definition.Dimension()
        view_model.key = dim_info.key
        view_model.id = dim_info.id
        view_model.name = dim_info.name
        view_model.isGeo = dim_info.is_geo
        view_model.datasetId = self.dataset

        for fields in dim_info.fields:
            view_model.fields.append(view_definition.Field(fields))

        frame_data = []
        headers = ['key', 'name', 'level', 'parent key', 'parent id', 'parent name', 'hasdata']
        header_filled = False

        parents = {}
        for item in dim_info.items:
            parents[item.level] = {
                'id': item.fields['id'] if 'id' in item.fields else None,
                'key': item.key,
                'name': item.name
            }

            parent = {'id': None, 'key': -1, 'name': None}
            if item.level > 0:
                parent = parents[item.level - 1]

            row = [item.key, item.name, item.level, parent['key'], parent['id'], parent['name'], item.hasdata]
            for field in dim_info.fields:
                if not header_filled:
                    headers.append(field['name'])

                if field['name'] in item.fields:
                    row.append(item.fields[field['name']])
                else:
                    row.append(None)

            if not header_filled:
                header_filled = True
            frame_data.append(row)
        
        view_model.members = pandas.DataFrame(frame_data, columns=headers)

        return view_model



import pytest
import knoema


@pytest.mark.parametrize('dataset_id', ['IMFWEO2020Apr', 'CDIACTACHIINDUSAA'])
def test_metadata(get_client, dataset_id):
    assert knoema.dataset(dataset_id)['id'] == dataset_id



"""This is main package module"""

from knoema.api_config import ApiConfig
from knoema.api_client import ApiClient
from knoema.data_reader import MnemonicsDataReader, StreamingDataReader, TransformationDataReader
from knoema.data_reader import DimensionMetadataReader
from knoema.api_definitions import is_equal_strings_ignore_case
from knoema.api_definitions_sema import Company
from knoema.api_definitions_search import SearchResults
from knoema.upload_frame import FrameTransformerFactory, FileLayerWrapper

def dataset(id):
    """Use this function to get dataset metadata."""

    config = ApiConfig()
    client = ApiClient(config.host, config.app_id, config.app_secret)
    client.check_correct_host()

    ds = client.get_dataset_meta(id)
    res = {}
    res.update(ds.data)

    range = client.get_daterange(id)
    res['startDate'] = range.start_date
    res['endDate'] = range.end_date
    res['frequencies'] = range.frequencies

    return res

def dimension(dataset, dimension):
    """Use this function to get dimension metadata"""

    config = ApiConfig()
    client = ApiClient(config.host, config.app_id, config.app_secret)
    client.check_correct_host()

    if not dataset:
        raise ValueError('Dataset id is not specified')

    if not dimension:
        raise ValueError('Dimension id is not specified')

    reader = DimensionMetadataReader(client, dataset, dimension)

    return reader.get()

def get(dataset = None, include_metadata = False, mnemonics = None, transform = None, separator = None, group_by = None, columns = None, **dim_values):
    """Use this function to get data from Knoema dataset."""

    if not dataset and not mnemonics:
        raise ValueError('Dataset id is not specified')

    config = ApiConfig()
    client = ApiClient(config.host, config.app_id, config.app_secret)
    client.check_correct_host()

    ds = client.get_dataset(dataset) if dataset else None

    if columns is not None and isinstance(columns, str):
        columns = columns.split(';')

    frequency = None
    time = None
    time_params_count = 0
    has_agg = False
    advanced = False
    for name, value in dim_values.copy().items():
        if is_equal_strings_ignore_case(name, 'frequency'):
            frequency = value
            del dim_values[name]
            continue

        if is_equal_strings_ignore_case(name, 'transform'):
            transform = value
            del dim_values[name]
            continue

        if is_equal_strings_ignore_case(name, 'timerange'):
            time = value
            del dim_values[name]
            if time != None: 
                dim_values['timerange'] = time
                ++time_params_count
            continue

        if is_equal_strings_ignore_case(name, 'timesince'):
            time = value
            del dim_values[name]
            if time != None: 
                dim_values['timesince'] = time
                advanced = True
                ++time_params_count
            continue

        if is_equal_strings_ignore_case(name, 'timelast'):
            time = value
            del dim_values[name]
            if time != None: 
                dim_values['timelast'] = time
                advanced = True
                ++time_params_count
            continue
        
        if is_equal_strings_ignore_case(name, 'timemembers'):
            time = value
            del dim_values[name]
            if time != None: 
                dim_values['timemembers'] = time
                advanced = True
                ++time_params_count
            continue

        if len(value) > 0 and value[0].startswith('@'):
            has_agg = True

    if time_params_count > 1:
        raise ValueError('Only one parameter should be passed: timerange, timesince, timelast, timemebers')
    
    if advanced and (group_by or columns):
        raise ValueError('Advanced time modes and multiple frequencies can\'t be used with group_by or columns parameters')
    
    if mnemonics:
        reader =  MnemonicsDataReader(client, mnemonics, transform, frequency)
        reader.columns = columns
        reader.include_metadata = include_metadata
        reader.dataset = ds

        if separator:
            reader.separator = separator
            
        return reader.get_pandasframe()
    
    if not dataset:
        raise ValueError('Dataset id is not specified')

    if ds.type == 'Regular' and not has_agg and group_by:
        if frequency != None: dim_values['frequency'] = frequency
        metadata_reader = StreamingDataReader(client, dim_values)
        metadata_reader.columns = columns
        metadata_reader.dataset = ds
        if separator:
            metadata_reader.separator = separator

        metadata = metadata_reader.get_series_metadata()

        reader = StreamingDataReader(client, None, transform, group_by)
        reader.columns = columns
        reader.include_metadata = include_metadata
        reader.dataset = ds

        return reader.get_pandasframe_by_metadata_grouped(metadata, frequency, time)

    if ds.type == 'Regular' and frequency != None:
        dim_values['frequency'] = frequency

    if ds.type == 'Regular' and columns != None:
        reader = StreamingDataReader(client, dim_values, transform)
    else:
        reader = TransformationDataReader(client, dim_values, transform, frequency)

    reader.columns = columns
    reader.include_metadata = include_metadata
    reader.dataset = ds

    if separator:
        reader.separator = separator
            
    return reader.get_pandasframe()

def ticker(ticker):
    """Use this function to get data about company"""

    if not ticker:
        raise ValueError('Ticker or company name is not specified')

    config = ApiConfig()
    client = ApiClient(config.host, config.app_id, config.app_secret)
    client.check_correct_host()

    company_int = client.get_company_info(ticker)

    return Company(company_int, client)

def search(query):
    """Use this function to make search request"""

    if not query:
        raise ValueError('Query is not specified')

    config = ApiConfig()
    client = ApiClient(config.host, config.app_id, config.app_secret)
    client.check_correct_host()

    search_results_int = client.search(query)
    search_results = SearchResults(search_results_int, client)

    if search_results.instant != None:
        if search_results.instant.type == 'ConceptBind':
            company_int = client.get_company_info(search_results.instant.id)

            search_results.instant.company = Company(company_int, client)

    return search_results

def upload(file_path_or_frame, dataset=None, public=False, name = None):
    """Use this function to upload data to Knoema dataset."""

    config = ApiConfig()
    client = ApiClient(config.host, config.app_id, config.app_secret)

    if isinstance(file_path_or_frame, str):
        return client.upload(file_path_or_frame, dataset, public, name)

    frame_transformer = FrameTransformerFactory(file_path_or_frame).get_transformer()

    with FileLayerWrapper() as fw:
        file = frame_transformer.prepare(fw, dataset, name)
        return client.upload(file, dataset, public, name)

def delete(dataset):
    """Use this function to delete dataset by it's id."""
    
    config = ApiConfig()
    client = ApiClient(config.host, config.app_id, config.app_secret)
    client.check_correct_host()
    client.delete(dataset)
    return ('Dataset {} has been deleted successfully'.format(dataset))

def verify(dataset, publication_date, source, refernce_url):
    """Use this function to verify a dataset."""

    config = ApiConfig()
    client = ApiClient(config.host, config.app_id, config.app_secret)
    client.check_correct_host()
    client.verify(dataset, publication_date, source, refernce_url)
    


"""This is test module for knoema client (Upload)"""
"""Works only with special credentials"""

import unittest
import datetime
import knoema
import urllib
import pandas
import os
import numpy

class TestKnoemaClient(unittest.TestCase):
    """This is class with knoema client unit tests (Upload)"""

    base_host = 'knoema.com'

    def setUp(self):
        apicfg = knoema.ApiConfig()
        apicfg.host = os.environ['BASE_HOST'] if 'BASE_HOST' in os.environ else self.base_host
        apicfg.app_id = os.environ['KNOEMA_APP_ID'] if 'KNOEMA_APP_ID' in os.environ else ''
        apicfg.app_secret = os.environ['KNOEMA_APP_SECRET'] if 'KNOEMA_APP_SECRET' in os.environ else ''

    def test_delete_dataset_negative(self):
        """The method is negative test on dataset deletion"""
         
        with self.assertRaises(urllib.error.HTTPError) as context:
            knoema.delete('non_existing_id')
        self.assertTrue('HTTP Error 400: Bad Request' in str(context.exception))

    def test_verify_dataset_negative(self):
        """The method is negative test on dataset verification"""

        with self.assertRaises(ValueError) as context:
            knoema.verify('non_existing_id', datetime.date.today(), 'IMF', 'http://knoema.com/')
        self.assertTrue("Dataset has not been verified, because of the following error(s): Requested dataset doesn't exist or you don't have access to it." in str(context.exception))

    def test_incorrect_host_delete_dataset(self):
        """The method is negative test on delete dataset with incorrect host"""

        with self.assertRaises(ValueError) as context:
            apicfg = knoema.ApiConfig()
            apicfg.host = 'knoema_incorect.com'
            knoema.delete('dataset')
        self.assertTrue("The specified host knoema_incorect.com does not exist" in str(context.exception))

    def test_incorrect_host_verify_dataset(self):
        """The method is negative test on verify dataset with incorrect host"""

        with self.assertRaises(ValueError) as context:
            apicfg = knoema.ApiConfig()
            apicfg.host = 'knoema_incorect.com'
            knoema.verify('non_existing_id', datetime.date.today(), 'IMF', 'http://knoema.com')
        self.assertTrue("The specified host knoema_incorect.com does not exist" in str(context.exception))    

    def test_upload_generated_frames(self):
        tuples = list(zip(*[['bar', 'bar', 'baz', 'baz', 'foo', 'foo', 'qux', 'qux'], ['one', 'two', 'one', 'two', 'one', 'two', 'one', 'two']]))
        index = pandas.MultiIndex.from_tuples(tuples, names=['first', 'second'])
        dates = pandas.date_range('1/1/2000', periods=8)
        frame = pandas.DataFrame(numpy.random.randn(8, 8), index=dates, columns=index)
        res = knoema.upload(frame, name = 'Test dataset')
        self.assertIs(type(res), str)
        self.assertEqual(len(res), 7)

        frame = pandas.DataFrame(numpy.random.randn(8, 4), index=dates, columns=['A', 'B', 'C', 'D'])
        res = knoema.upload(frame, name = 'Test dataset')
        self.assertIs(type(res), str)
        self.assertEqual(len(res), 7)

    def test_upload_frames_from_existing_datasets(self):
        frame = knoema.get('xmhdwqf', company='UBER', indicator='Annual', frequency='A', timerange='2018-2020')
        res = knoema.upload(frame, name = 'Test dataset')

        self.assertIs(type(res), str)
        self.assertEqual(len(res), 7)